"""
Contains functions that are useful for describing instances in a consistent manner
"""
import asyncio
from typing import Any
from typing import Dict
from typing import Tuple


async def convert_to_present(
    hub, ctx, describe_instances: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Convert instances from ec2.describe_instances() to aws.ec2.instance.present states

    This is the preferred "meta" function for collecting information about multiple instances
    """
    result = {}
    coros = []
    for reservation in describe_instances.get("Reservations", []):
        for instance in reservation["Instances"]:
            instance_id = instance["InstanceId"]
            placement = instance.get("Placement", {})
            metadata_options = instance.get("MetadataOptions", {})
            private_dns_name_options = instance.get("PrivateDnsNameOptions", {})
            result[instance_id] = dict(
                name=instance_id,
                resource_id=instance_id,
                image_id=instance.get("ImageId"),
                instance_type=instance.get("InstanceType"),
                volume_attachments=None,
                block_device_mappings=instance.get("BlockDeviceMappings"),
                ebs_optimized=instance.get("EbsOptimized"),
                kernel_id=instance.get("KernelId"),
                subnet_id=instance.get("SubnetId"),
                network_interfaces=instance.get("NetworkInterfaces"),
                monitoring_enabled=instance.get("Monitoring", {}).get("State")
                == "enabled",
                root_device_name=instance.get("RootDeviceName"),
                client_token=instance.get("ClientToken"),
                product_codes=instance.get("ProductCodes"),
                source_dest_check=instance.get("SourceDestCheck"),
                running=instance.get("State", {}).get("Name") == "running",
                private_ip_address=instance.get("PrivateIpAddress"),
                reservation_id=reservation.get("ReservationId"),
                owner_id=reservation.get("OwnerId"),
                availability_zone=placement.get("AvailabilityZone"),
                affinity=placement.get("Affinity"),
                group_name=placement.get("GroupName"),
                partition_number=placement.get("PartitionNumber"),
                host_id=placement.get("HostId"),
                tenancy=placement.get("Tenancy"),
                spread_domain=placement.get("SpreadDomain"),
                host_resource_group_arn=placement.get("HostResourceGroupArn"),
                user_data=None,
                disable_api_termination=None,
                ram_disk_id=instance.get("RamdiskId"),
                tags={tag["Key"]: tag["Value"] for tag in instance.get("Tags", [])},
                iam_profile_arn=instance.get("IamInstanceProfile", {}).get("Arn"),
                instance_initiated_shutdown_behavior=None,
                elastic_inference_accelerators=instance.get(
                    "ElasticInferenceAcceleratorAssociations"
                ),
                auto_recovery_enabled=not (
                    instance.get("MaintenanceOptions", {}).get("AutoRecovery")
                    == "disabled"
                ),
                sriov_net_support=instance.get("SriovNetSupport"),
                key_name=instance.get("KeyName"),
                elastic_gpu_specificatiosn=instance.get("ElasticGpuAssociations"),
                nitro_enclave_enabled=instance.get("EnclaveOptions", {}).get("Enabled"),
                license_arns=[
                    license_["LicenseConfigurationArn"]
                    for license_ in instance.get("Licenses", [])
                ],
                hibernation_enabled=instance.get("HibernationOptions", {}).get(
                    "Configured"
                ),
                market_type=None,
                max_price=None,
                spot_instance_type=None,
                block_duration_minutes=None,
                valid_until=None,
                instance_interruption_behavior=None,
                cpu_credits=None,
                cpu_core_count=instance.get("CpuOptions", {}).get("CoreCount"),
                cpu_threads_per_core=instance.get("CpuOptions", {}).get(
                    "ThreadsPerCore"
                ),
                http_tokens=metadata_options.get("HttpTokens"),
                http_put_response_hop_limit=metadata_options.get(
                    "HttpPutResponseHopLimit"
                ),
                http_endpoint_enabled=metadata_options.get("HttpEndpoint") == "enabled",
                http_protocol_ipv6_enabled=metadata_options.get("HttpProtocolIpv6")
                == "enabled",
                metadata_tags_enabled=metadata_options.get("InstanceMetadataTags")
                == "enabled",
                hostname_type=private_dns_name_options.get("HostnameType"),
                enable_resource_name_dns_a_record=private_dns_name_options.get(
                    "EnableResourceNameDnsARecord"
                ),
                enable_resource_name_dns_aaaa_record=private_dns_name_options.get(
                    "EnableResourceNameDnsAAAARecord"
                ),
                capacity_reservation_preference=instance.get(
                    "CapacityReservationSpecification", {}
                ).get("CapacityReservationPreference"),
                bootstrap=[],
            )

            coros.append(
                # launch_template config get most parameters for an instance
                hub.tool.aws.ec2.instance.state.config(ctx, instance_id=instance_id)
            )

    # This can be a heavy process if there are many instances, use coroutines to collect them all at simultaneously
    for ret in asyncio.as_completed(coros):
        instance_id, config = await ret
        result[instance_id].update(config)

    return hub.tool.aws.ec2.instance.data.sanitize_dict(result)


async def config(hub, ctx, *, instance_id: str) -> Tuple[str, Dict[str, Any]]:
    """
    Retrieves the configuration data of the specified instance.

    This action calls on other describe actions to get instance information.
    Depending on your instance configuration, you may need to allow the following actions in your IAM policy:
        - DescribeSpotInstanceRequests
        - DescribeInstanceCreditSpecifications
        - DescribeVolumes
        - DescribeInstanceAttribute
        - DescribeElasticGpus.
    Or, you can allow describe* depending on your instance requirements.
    """
    config = {}
    response = await hub.exec.boto3.client.ec2.get_launch_template_data(
        ctx, InstanceId=instance_id
    )
    if response:
        hub.log.trace(
            f"Collecting instance '{instance_id}' config from launch template data"
        )
        # This is ideal and concise
        launch_config = hub.tool.aws.ec2.instance.state.parse_launch_template_data(
            **response.ret["LaunchTemplateData"]
        )
        for key, value in launch_config.items():
            if config.get(key):
                continue
            config[key] = value
    else:
        # Maybe we lack permissions, or we are using localstack and get_launch_template_data is not implemented yet
        # Get that same information elsewhere
        hub.log.trace(f"Collecting instance '{instance_id}' config manually")
        attributes_ = await hub.tool.aws.ec2.instance.state.attributes(
            ctx, instance_id=instance_id
        )
        for key, value in attributes_.items():
            if config.get(key):
                continue
            config[key] = value

    # Collect information that was not part of the launch template
    extended_attributes_ = await hub.tool.aws.ec2.instance.state.extended_attributes(
        ctx, instance_id=instance_id
    )
    for key, value in extended_attributes_.items():
        if config.get(key):
            continue
        config[key] = value

    return instance_id, config


def parse_launch_template_data(hub, **launch_config) -> Dict[str, Any]:
    """
    Parse LaunchTemplateData to collect information about a single instance
    """
    placement = launch_config.get("Placement", {})
    instance_market_options = launch_config.get("InstanceMarketOptions", {})
    spot_options = instance_market_options.get("SpotOptions", {})
    metadata_options = launch_config.get("MetadataOptions", {})
    private_dns_name_options = launch_config.get("PrivateDnsNameOptions", {})

    tags = {}
    for tag_spec in launch_config.get("TagSpecifications", []):
        if tag_spec["ResourceType"] == "instance":
            for tag in tag_spec["Tags"]:
                tags[tag["Key"]] = tag["Value"]

    return dict(
        image_id=launch_config.get("ImageId"),
        instance_type=launch_config.get("InstanceType"),
        # block_device_mappings
        ebs_optimized=launch_config.get("EbsOptimized"),
        kernel_id=launch_config.get("KernelId"),
        # subnet_id=None,
        network_interfaces=launch_config.get("NetworkInterfaces"),
        monitoring_enabled=launch_config.get("Monitoring", {}).get("Enabled"),
        # root_device_name
        # product_codes
        # source_dest_check
        # running
        # private_ip_address
        # reservation_id
        # owner_id
        availability_zone=placement.get("AvailabilityZone"),
        affinity=placement.get("Affinity"),
        group_name=placement.get("GroupName"),
        partition_number=placement.get("PartitionNumber"),
        host_id=placement.get("HostId"),
        tenancy=placement.get("Tenancy"),
        spread_domain=placement.get("SpreadDomain"),
        host_resource_group_arn=placement.get("HostResourceGroupArn"),
        user_data=launch_config.get("UserData"),
        disable_api_termination=launch_config.get("DisableApiTermination"),
        ram_disk_id=launch_config.get("RamDiskId"),
        tags=tags,
        # iam_profile_arn
        instance_initiated_shutdown_behavior=launch_config.get(
            "InstanceInitiatedShutdownBehavior"
        ),
        elastic_inference_accelerators=launch_config.get(
            "ElasticInferenceAccelerators"
        ),
        # auto_recovery_enabled
        # sriov_net_support
        # key_name
        elastic_gpu_specifications=launch_config.get("ElasticGpuSpecifications"),
        nitro_enclave_enabled=launch_config.get("EnclaveOptions", {}).get("Enabled"),
        hibernation_enabled=launch_config.get("HibernationOptions", {}).get(
            "Configured"
        ),
        market_type=instance_market_options.get("MarketType"),
        max_price=spot_options.get("MaxPrice"),
        spot_instance_type=spot_options.get("SpotInstanceType"),
        block_duration_minutes=spot_options.get("BlockDurationMinutes"),
        valid_until=str(spot_options.get("ValidUntil", "")),
        instance_interruption_behavior=spot_options.get("InstanceInterruptionBehavior"),
        cpu_credits=launch_config.get("CreditSpecification", {}).get("CpuCredits"),
        cpu_core_count=launch_config.get("CpuOptions", {}).get("CoreCount"),
        cpu_threads_per_core=launch_config.get("CpuOptions", {}).get("ThreadsPerCore"),
        http_tokens=metadata_options.get("HttpTokens"),
        http_put_response_hop_limit=metadata_options.get("HttpPutResponseHopLimit"),
        http_endpoint_enabled=metadata_options.get("HttpEndpoint") == "enabled",
        http_protocol_ipv6_enabled=metadata_options.get("HttpProtocolIpv6")
        == "enabled",
        metadata_tags_enabled=metadata_options.get("InstanceMetadataTags") == "enabled",
        hostname_type=private_dns_name_options.get("HostnameType"),
        enable_resource_name_dns_a_record=private_dns_name_options.get(
            "EnableResourceNameDnsARecord"
        )
        == "enabled",
        enable_resource_name_dns_aaaa_record=private_dns_name_options.get(
            "EnableResourceNameDnsAAAARecord"
        )
        == "enabled",
        capacity_reservation_preference=launch_config.get(
            "CapacityReservationSpecification", {}
        ).get("CapacityReservationPreference"),
        # bootstrap
    )


async def attributes(hub, ctx, *, instance_id: str) -> Dict[str, Any]:
    """
    Manually collect information about a single instance, these are also gathered from parsing launch template data
    """
    instance = {}

    ret = await hub.exec.boto3.client.ec2.describe_instance_attribute(
        ctx, Attribute="userData", InstanceId=instance_id
    )
    if ret:
        instance["user_data"] = ret.ret["UserData"]

    ret = await hub.exec.boto3.client.ec2.describe_instance_attribute(
        ctx, Attribute="disableApiTermination", InstanceId=instance_id
    )
    if ret:
        instance["disable_api_termination"] = ret.ret["DisableApiTermination"].get(
            "Value"
        )

    ret = await hub.exec.boto3.client.ec2.describe_instance_attribute(
        ctx, Attribute="instanceInitiatedShutdownBehavior", InstanceId=instance_id
    )
    if ret:
        instance["instance_initiated_shutdown_behavior"] = ret.ret[
            "InstanceInitiatedShutdownBehavior"
        ].get("Value")

    ret = await hub.exec.boto3.client.ec2.describe_instance_attribute(
        ctx, Attribute="enclaveOptions", InstanceId=instance_id
    )
    if ret:
        # This is not supported by AWS yet
        instance["enclave_options"] = ret.ret["EnclaveOptions"]["Enabled"]

    ret = await hub.exec.boto3.client.ec2.describe_instance_attribute(
        ctx, Attribute="ramdisk", InstanceId=instance_id
    )
    if ret:
        instance["ram_disk_id"] = ret.ret["RamdiskId"]

    ret = await hub.exec.boto3.client.ec2.describe_tags(
        ctx,
        Filters=[
            {"Name": "resource-id", "Values": [instance_id]},
        ],
    )
    if ret:
        instance["tags"] = {tag["Key"]: tag.get("Value") for tag in ret.ret["Tags"]}

    return instance


async def extended_attributes(hub, ctx, *, instance_id: str) -> Dict[str, Any]:
    """
    Manually collect information about a single instance
    These are not gathered from parsing launch template data or from describe_instances
    """
    instance = {}

    # iam_profile_arn
    ret = await hub.exec.boto3.client.ec2.describe_iam_instance_profile_associations(
        ctx, Filters=[{"Name": "instance-id", "Values": [instance_id]}]
    )
    if ret:
        instance["iam_profile_arn"] = {}
        for association in ret.ret["IamInstanceProfileAssociations"]:
            if association["State"] == "associating":
                # Prefer reporting what will soon be true
                instance["iam_profile_arn"] = association["IamInstanceProfile"]["Arn"]
                break
            elif association["State"] == "associated":
                # Fallback to the current configuration
                instance["iam_profile_arn"] = association["IamInstanceProfile"]["Arn"]
                break
            else:
                instance["iam_profile_arn"] = association["IamInstanceProfile"]["Arn"]
                # Don't break, this might be disassociated

    # sriov_net_support
    ret = await hub.exec.boto3.client.ec2.describe_instance_attribute(
        ctx, Attribute="sriovNetSupport", InstanceId=instance_id
    )
    if ret:
        instance["sriov_net_support"] = ret.ret["SriovNetSupport"].get("Value")

    # block_device_mappings
    volume_attachments = {}
    ret = await hub.exec.boto3.client.ec2.describe_volumes(
        ctx, Filters=[{"Name": "attachment.instance-id", "Values": [instance_id]}]
    )
    if ret:
        for volume in ret.ret.get("Volumes", ()):
            for attachment in volume.get("Attachments", ()):
                if attachment["InstanceId"] == instance_id:
                    device = attachment["Device"]
                    volume_attachments[device] = volume["VolumeId"]

    instance["volume_attachments"] = volume_attachments

    return instance


def test(hub, **kwargs) -> Dict[str, Any]:
    """
    Compute the state based on the parameters passed to an instance.present function for ctx.test
    """
    result = {}
    for k, v in kwargs.items():
        # Ignore kwargs that were None
        if v is None:
            continue
        result[k] = v
    return result



from typing import Any
from typing import Dict


def convert_to_present(
    hub, ctx, describe_network_interfaces: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Convert the Network Interface to present
    """
    result = {}

    for network_interface in describe_network_interfaces.get("NetworkInterfaces", ()):
        resource_id = network_interface.get("NetworkInterfaceId")

        result[resource_id] = dict(
            name=resource_id,
            resource_id=resource_id,
            attachment_id=network_interface.get("Attachment", {}).get("AttachmentId"),
        )

    return result



"""Exec module for managing IAM password policy."""
from typing import Dict


async def get(hub, ctx, name: str) -> Dict:
    """Retrieves the password policy for the Amazon Web Services account.

    Args:
        name(str): Password policy name.

    Returns:
        Dict[bool, list, dict or None]:

        result(bool):
            Whether the result of the function has been successful (``True``) or not (``False``).
        comment(list):
            A list of messages.
        ret(dict or None):
           The IAM password policy in "present" format.

    Examples:
        Calling this exec module function from the cli:

        .. code-block:: bash

            idem exec aws.iam.password_policy.get name="idem_name"

        Calling this exec module function from within a state module in pure python:

        .. code-block:: python

            async def state_function(hub, ctx, name, resource_id, **kwargs):
                ret = await hub.exec.aws.iam.password_policy.get(
                    ctx, name=name
                )

    """
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.exec.boto3.client.iam.get_account_password_policy(ctx)

    if not ret["result"]:
        if "NoSuchEntity" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.iam.password_policy", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    result[
        "ret"
    ] = hub.tool.aws.iam.conversion_utils.convert_raw_password_policy_to_present(
        ctx, raw_password_policy=ret["ret"]["PasswordPolicy"]
    )
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict

"""
Util function to convert raw resource state from AWS Launch Configuration to present input format.
"""


def convert_raw_launch_configuration_to_present(
    hub, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    resource_id = raw_resource.get("LaunchConfigurationName")
    resource_parameters = OrderedDict(
        {
            "LaunchConfigurationName": "name",
            "ImageId": "image_id",
            "KeyName": "key_name",
            "SecurityGroups": "security_groups",
            "ClassicLinkVPCId": "classic_link_vpc_id",
            "ClassicLinkVPCSecurityGroups": "classic_link_vpc_security_groups",
            "UserData": "user_data",
            "InstanceId": "instance_id",
            "InstanceType": "instance_type",
            "KernelId": "kernel_id",
            "RamdiskId": "ramdisk_id",
            "BlockDeviceMappings": "block_device_mappings",
            "InstanceMonitoring": "instance_monitoring",
            "SpotPrice": "spot_price",
            "IamInstanceProfile": "iam_instance_profile",
            "EbsOptimized": "ebs_optimized",
            "AssociatePublicIpAddress": "associate_public_ip_address",
            "PlacementTenancy": "placement_tenancy",
            "MetadataOptions": "metadata_options",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource and raw_resource.get(parameter_raw):
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    return resource_translated


def convert_raw_scaling_policy_to_present(
    hub, ctx, raw_resource: Dict[str, Any]
) -> Dict[str, Any]:
    resource_id = raw_resource.get("PolicyName")

    resource_parameters = OrderedDict(
        {
            "AutoScalingGroupName": "auto_scaling_group_name",
            "PolicyType": "policy_type",
            "AdjustmentType": "adjustment_type",
            "MinAdjustmentStep": "min_adjustment_step",
            "MinAdjustmentMagnitude": "min_adjustment_magnitude",
            "ScalingAdjustment": "scaling_adjustment",
            "Cooldown": "cooldown",
            "MetricAggregationType": "metric_aggregation_type",
            "EstimatedInstanceWarmup": "estimated_instance_warmup",
            "Enabled": "enabled",
        }
    )
    resource_translated = {
        "name": raw_resource.get("PolicyName"),
        "resource_id": resource_id,
    }
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource and raw_resource.get(parameter_raw):
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    if "StepAdjustments" in raw_resource:
        resource_translated["step_adjustments"] = raw_resource.get(
            "StepAdjustments"
        ).copy()

    if "TargetTrackingConfiguration" in raw_resource:
        resource_translated["target_tracking_configuration"] = raw_resource.get(
            "TargetTrackingConfiguration"
        ).copy()

    if "PredictiveScalingConfiguration" in raw_resource:
        resource_translated["predictive_scaling_configuration"] = raw_resource.get(
            "PredictiveScalingConfiguration"
        ).copy()

    return resource_translated



"""Exec module for managing VPC Peering Connection Options."""
from typing import Dict
from typing import List


async def get(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    filters: List = None,
) -> Dict:
    """Get EC2 vpc peering connection options from AWS account.

    Use an un-managed VPC peering connection options as a data-source.
    Supply one of the inputs as the filter.

    Args:
        name(str):
            The name of the Idem state.

        resource_id(str, Optional):
            AWS VPC peering connection id to identify the resource.

        filters(list, Optional):
            One or more filters: for example, tag :<key>, tag-key.
            A complete list of filters can be found at
            https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_vpc_peering_connections

    Examples:
        Calling this exec module function from the cli with resource_id

        .. code-block:: bash

            idem exec aws.ec2.vpc_peering_connection_options.get resource_id="resource_id" name="name"


        Calling this exec module function from the cli with filters.
        The filters can be passed as a json string that will be parsed into a python structure by idem's config loader.

        .. code-block:: bash

            idem exec aws.ec2.vpc_peering_connection_options.get filters='[{"name": "filter_name", "values": ["filter_value1"]}]'

        Calling this exec module function from within a state module in pure python

        .. code-block:: python

            async def state_function(hub, ctx, name, resource_id, **kwargs):
                ret = await hub.exec.aws.ec2.vpc_peering_connection_options.get(ctx, resource_id=resource_id, name=name, **kwargs)
    """
    result = await hub.tool.aws.ec2.vpc_peering_connection_utils.get_vpc_peering_connection_raw(
        ctx, name, resource_id, "aws.ec2.vpc_peering_connection", filters
    )
    raw_resource = result["ret"]

    result[
        "ret"
    ] = hub.tool.aws.ec2.conversion_utils.convert_raw_vpc_peering_connection_options_to_present(
        raw_resource=raw_resource, idem_resource_name=name
    )

    return result



"""State module for managing EC2 VPC Peering Connection."""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    peer_owner_id: str = None,
    peer_vpc_id: str = None,
    vpc_id: str = None,
    peer_region: str = None,
    tags: Dict[str, str] = None,
    status: str = None,
) -> Dict[str, Any]:
    """Requests to create or update a VPC peering connection.

    Requests a VPC peering connection between two VPCs: a requester VPC that you own and an accepter VPC with which
    to create the connection. The accepter VPC can belong to another Amazon Web Services account and can be in a
    different Region to the requester VPC. The requester VPC and accepter VPC cannot have overlapping CIDR blocks.
    Limitations and rules apply to a VPC peering connection. For more information, see the limitations section in
    the VPC Peering Guide.  The owner of the accepter VPC must accept the peering request to activate the peering
    connection. The VPC peering connection request expires after 7 days, after which it cannot be accepted or
    rejected. If you create a VPC peering connection request between VPCs with overlapping CIDR blocks, the VPC
    peering connection has a status of failed.

    NOTE: These five parameters - peer_owner_id, peer_vpc_id, vpc_id, peer_region and status
    can't be updated for a given VPC peering connection.
    The only thing that could be updated is the tags,
    and when status is set to active when a new resource is created in real,
    an attempt will be made to set the vpc peering connection status to active.
    In case of an update attempt of the previously mentioned five parameters,
    where resource_id is passed for an existing connection, they will be ignored.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

        peer_owner_id(str, Optional):
            The Amazon Web Services account ID of the owner of the accepter VPC.
            Default: Your Amazon Web Services account ID. Defaults to None.

        peer_vpc_id(str, Optional):
            The ID of the VPC with which you are creating the VPC peering connection.
            You must specify this parameter in the request. Defaults to None.

        vpc_id(str, Optional):
            The ID of the requester VPC. You must specify this parameter in the request.
            Defaults to None.

        peer_region(str, Optional):
            The Region code for the accepter VPC, if the accepter VPC is located in a Region
            other than the Region in which you make the request.
            Default: The Region in which you make the request. Defaults to None.

        tags(Dict, Optional):
            Dict in the format of {tag-key: tag-value} The tags to assign to the peering connection.
            Each tag consists of a key name and an associated value. Defaults to None.

        status: (str, Optional)
            The current status of the vpc peering connection.

    Request Syntax:
       .. code-block:: sls

            [vpc-peering-connection-id]:
              aws.ec2.vpc_peering_connection.present:
              - resource_id: "string"
              - name: "string"
              - peer_owner_id: "string"
              - peer_region: "string"
              - peer_vpc_id: "string"
              - vpc_id: "string"
              - tags: "Dict"
              - status: "string"

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_present:
              aws.ec2.vpc_peering_connection.present:
                - resource_id: pcx-ae89ce9b
                - name: pcx-ae89ce9b
                - peer_owner_id: '000000000000'
                - peer_region: us-west-2
                - peer_vpc_id: vpc-98c058ae
                - vpc_id: vpc-2c90d746
                - status: active
                - tags:
                    first_key: first_value
                    second_key: second_value
                    third_key: third_value
                    fourth_key: fourth_value
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource_updated = False
    if resource_id:
        before = await hub.exec.aws.ec2.vpc_peering_connection.get(
            ctx, name=name, resource_id=resource_id
        )
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result

        result["old_state"] = before["ret"]
        plan_state = copy.deepcopy(result["old_state"])

        # Update tags
        if tags is not None and tags != result["old_state"].get("tags"):
            update_ret = await hub.exec.aws.ec2.tag.update_tags(
                ctx=ctx,
                resource_id=resource_id,
                old_tags=result["old_state"].get("tags"),
                new_tags=tags,
            )

            result["result"] = update_ret["result"]
            result["comment"] = update_ret["comment"]
            if not update_ret["result"]:
                return result

            resource_updated = resource_updated or bool(update_ret["ret"])
            if update_ret["ret"] is not None and ctx.get("test", False):
                plan_state["tags"] = update_ret["ret"]

        if resource_updated:
            if ctx.get("test", False):
                result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                    "aws.ec2.vpc_peering_connection", name
                )
            else:
                result["comment"] += hub.tool.aws.comment_utils.update_comment(
                    "aws.ec2.vpc_peering_connection", name
                )
        else:
            result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
                "aws.ec2.vpc_peering_connection", name
            )
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "peer_owner_id": peer_owner_id,
                    "peer_vpc_id": peer_vpc_id,
                    "vpc_id": vpc_id,
                    "peer_region": peer_region,
                    "tags": tags,
                    "status": status,
                },
            )

            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                "aws.ec2.vpc_peering_connection", name
            )
            return result

        tag_specifications = [
            {
                "ResourceType": "vpc-peering-connection",
                "Tags": hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags),
            }
        ]

        ret = await hub.exec.boto3.client.ec2.create_vpc_peering_connection(
            ctx,
            **{
                "PeerOwnerId": peer_owner_id,
                "PeerVpcId": peer_vpc_id,
                "VpcId": vpc_id,
                "PeerRegion": peer_region,
                "TagSpecifications": tag_specifications,
            },
        )

        result["new_state"] = {"name": name, "resource_id": resource_id}
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] += ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            "aws.ec2.vpc_peering_connection", name
        )

        new_vpc_peering_connection = ret["ret"]["VpcPeeringConnection"]
        resource_id = new_vpc_peering_connection["VpcPeeringConnectionId"]

        if status == "active":
            accept_ret = await hub.exec.boto3.client.ec2.accept_vpc_peering_connection(
                ctx, VpcPeeringConnectionId=resource_id
            )

            if not accept_ret["result"]:
                result["comment"] += accept_ret["comment"]
                result["result"] = False
                return result

            new_status = accept_ret["ret"]["VpcPeeringConnection"]["Status"]["Code"]
            result[
                "comment"
            ] += hub.tool.aws.comment_utils.resource_status_updated_comment(
                "aws.ec2.vpc_peering_connection", name, new_status
            )

            if not accept_ret["result"]:
                result["result"] = False
                return result

    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif (not result["old_state"]) or resource_updated:
        after = await hub.exec.aws.ec2.vpc_peering_connection.get(
            ctx=ctx, name=name, resource_id=resource_id
        )
        if not after["result"]:
            result["result"] = False
            result["comment"] = after["comment"]
            return result

        result["new_state"] = copy.deepcopy(after["ret"])
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])

    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    """Deletes a VPC peering connection.

    Either the owner of the requester VPC or the owner of the accepter VPC can
    delete the VPC peering connection if it's in the active state. The owner of the requester VPC can delete a VPC
    peering connection in the pending-acceptance state. You cannot delete a VPC peering connection that's in the
    failed state.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str, Optional):
            An identifier of the resource in the provider.

    Request Syntax:
       .. code-block:: sls

            [vpc-peering-connection-id]:
              aws.ec2.vpc_peering_connection.absent:
              - resource_id: "string"
              - name: "string"

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_absent:
              aws.ec2.vpc_peering_connection.absent:
                - name: value
                - resource_id: value
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            "aws.ec2.vpc_peering_connection", name
        )
        return result

    before = await hub.exec.aws.ec2.vpc_peering_connection.get(
        ctx=ctx, name=name, resource_id=resource_id
    )
    if not before["result"]:
        result["result"] = False
        result["comment"] = before["comment"]
        return result
    if not before["ret"] or before["ret"]["status"] == "deleted":
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            "aws.ec2.vpc_peering_connection", name
        )
    else:
        result["old_state"] = before["ret"]

        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                "aws.ec2.vpc_peering_connection", name
            )
        else:
            # Deleting the vpc peering connection
            ret = await hub.exec.boto3.client.ec2.delete_vpc_peering_connection(
                ctx, **{"VpcPeeringConnectionId": resource_id}
            )

            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result

            result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                "aws.ec2.vpc_peering_connection", name
            )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Describes one or more of your VPC peering connections.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws.ec2.vpc_peering_connection
    """
    result = {}
    ret = await hub.exec.boto3.client.ec2.describe_vpc_peering_connections(ctx)

    if not ret["result"]:
        hub.log.debug(
            f"Could not describe aws.ec2.vpc_peering_connection {ret['comment']}"
        )
        return {}

    for vpc_peering_connection in ret["ret"]["VpcPeeringConnections"]:
        resource_id = vpc_peering_connection.get("VpcPeeringConnectionId")
        vpc_peering_connection_translated = hub.tool.aws.ec2.conversion_utils.convert_raw_vpc_peering_connection_to_present(
            raw_resource=vpc_peering_connection, idem_resource_name=resource_id
        )

        result[vpc_peering_connection_translated["resource_id"]] = {
            "aws.ec2.vpc_peering_connection.present": [
                vpc_peering_connection_translated
            ]
        }

    return result



from typing import Dict
from typing import List


async def apply(
    hub,
    ctx,
    resource,
    *,
    old_value: Dict[str, str],
    new_value: Dict[str, str],
    comments: List[str],
) -> bool:
    """
    Modify an ec2 instance based on a single parameter in it's "present" state

      - volume_attachments:
          /dev/xvda: vol-0f0c6dac5d13b9dc4

    Args:
        hub:
        ctx: The ctx from a state module call
        resource: An ec2 instance resource object
        old_value: The previous value from the attributes of an existing instance
        new_value: The desired value from the ec2 instance present state parameters
        comments: A running list of comments abound the update process
    """
    # Attach new volumes
    # If a device name is in the new mappings, but not the old, its a new volume to attach
    volumes_to_attach = set(new_value.keys()) - set(old_value.keys())
    for device_name in volumes_to_attach:
        volume_id = new_value[device_name]
        ret = await hub.boto3.client.ec2.attach_volume(
            ctx, Device=device_name, InstanceId=resource.id, VolumeId=volume_id
        )
        if ret.comment:
            comments.append(ret.comment)
        if not ret.result:
            comments.append(f"Could not attach volume: {volume_id} to instance")
            return False

    # Remove old volumes
    # If a device name is in the old mappings, but not the new, its a volume to detach
    volumes_to_remove = set(old_value.keys()) - set(new_value.keys())
    for device_name in volumes_to_remove:
        volume_id = old_value[device_name]
        ret = await hub.boto3.client.ec2.detach_volume(
            ctx, Device=device_name, InstanceId=resource.id, VolumeId=volume_id
        )
        if ret.comment:
            comments.append(ret.comment)
        if not ret.result:
            comments.append(f"Could not detach volume: {volume_id} from instance")
            return False

    # If a volume exists in both places then check if the volume id has changed
    volumes_to_move = set(old_value.keys()).intersection(set(new_value.keys()))
    for device_name in volumes_to_move:
        if old_value[device_name] != new_value[device_name]:
            # Detach the old volume
            volume_id = old_value[device_name]
            ret = await hub.boto3.client.ec2.detach_volume(
                ctx, Device=device_name, InstanceId=resource.id, VolumeId=volume_id
            )
            if ret.comment:
                comments.append(ret.comment)
            if not ret.result:
                comments.append(f"Could not detach volume: {volume_id} from instance")
                return False

            # Attach the new volume
            volume_id = new_value[device_name]
            ret = await hub.boto3.client.ec2.attach_volume(
                ctx, Device=device_name, InstanceId=resource.id, VolumeId=volume_id
            )
            if ret.comment:
                comments.append(ret.comment)
            if not ret.result:
                comments.append(f"Could not attach volume: {volume_id} to instance")
                return False

    return True



"""State module for managing Amazon Organizations Policies."""
import copy
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]

TREQ = {
    "absent": {
        "require": [
            "aws.organizations.policy_attachment.absent",
        ],
    },
    "present": {
        "require": [
            "aws.organizations.account.present",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    description: str = None,
    policy_type: str = None,
    content: str = None,
    tags: List[Dict[str, Any]] or Dict[str, Any] = None,
) -> Dict[str, Any]:
    """Creates Organization Policy.

    Creates a policy of a specified type that you can attach to a root, an organizational unit (OU), or an individual
    AWS account.

    Args:
        name(str):
            The name of the policy.

        resource_id(str, Optional):
            The ID of the policy in Amazon Web Services.

        description(str, Optional):
            A description to assign to the policy.

        policy_type(str, Optional):
            The type of policy to create. Only supported values are
            SERVICE_CONTROL_POLICY and TAG_POLICY.

        content(str, Optional):
            The policy text content to add to the new policy. The text that you supply must adhere
            to the rules of the policy type.

        tags(dict or list, Optional):
            Dict in the format of {tag-key: tag-value} or List of tags in the format of
            [{"Key": tag-key, "Value": tag-value}] to associate with the policy.

            * Key (str, optional): The key identifier, or name, of the tag.
            * Value (str, optional): The string value that's associated with the key of the tag.

    Request syntax:
        .. code-block:: sls

            [idem_test_aws_organizations_policy]:
              aws.organizations.policy.present:
                - name: 'string'
                - resource_id: 'string'
                - description: 'string'
                - policy_type: 'SERVICE_CONTROL_POLICY|TAG_POLICY'
                - content: 'string'
                - tags:
                    'string': 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem_test_aws_organizations_policy:
              aws.organizations.policy.present:
                - name: 'idem_test_policy'
                - description: 'Enables admins of attached accounts to delegate all S3 permissions'
                - policy_type: 'SERVICE_CONTROL_POLICY'
                - content:
                    Version: '2012-10-17'
                    Statement:
                      - Sid: 'AllowAllS3Actions'
                        Effect: 'Allow'
                        Action: ['s3:*']
                - tags:
                    provider: idem
    """
    result = dict(comment=(), name=name, result=True, old_state=None, new_state=None)
    before = None

    if resource_id:
        before = await hub.exec.aws.organizations.policy.get(
            ctx=ctx, name=name, resource_id=resource_id
        )
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result

    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)
    change_dict = dict(Name=None, Description=None, Content=None)

    resource_updated = False

    if before:
        # Policy exists, update
        if not ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
                resource_type="aws.organizations.policy", name=name
            )

        result["old_state"] = copy.deepcopy(before["ret"])
        plan_state = copy.deepcopy(result["old_state"])

        resource_id = before["ret"]["resource_id"]

        try:

            if before["ret"]["name"] != name:
                change_dict["Name"] = name

            if before["ret"]["description"] != description:
                change_dict["Description"] = description

            if before["ret"]["content"] != content:
                change_dict["Content"] = content

            if any(value is not None for value in change_dict.values()):

                if ctx.get("test", False):
                    plan_state["resource_id"] = resource_id
                    if change_dict.get("Name"):
                        plan_state["name"] = change_dict.get("Name")
                    if change_dict.get("Description"):
                        plan_state["description"] = change_dict.get("Description")
                    if change_dict.get("Content"):
                        plan_state["content"] = change_dict.get("Content")
                    result[
                        "comment"
                    ] += hub.tool.aws.comment_utils.would_update_comment(
                        resource_type="aws.organizations.policy", name=name
                    )
                else:
                    update_policy_ret = (
                        await hub.exec.boto3.client.organizations.update_policy(
                            ctx,
                            PolicyId=resource_id,
                            Name=change_dict.get("Name"),
                            Description=change_dict.get("Description"),
                            Content=change_dict.get("Content"),
                        )
                    )

                    if not update_policy_ret["result"]:
                        result["comment"] = (
                            result["comment"] + update_policy_ret["comment"]
                        )
                        result["result"] = update_policy_ret["result"]
                        return result
                    resource_updated = True
                    result["comment"] = result["comment"] + (
                        f"Updated aws.organizations.policy '{name}'",
                    )

            # Update tags
            if (tags is not None) and tags != result["old_state"].get("tags"):
                update_ret = await hub.tool.aws.organizations.tag.update_tags(
                    ctx,
                    resource_id=resource_id,
                    old_tags=result["old_state"].get("tags"),
                    new_tags=tags,
                )
                result["result"] = result["result"] and update_ret["result"]
                result["comment"] = result["comment"] + update_ret["comment"]
                resource_updated = resource_updated or bool(update_ret["result"])
                if ctx.get("test", False) and update_ret["ret"] is not None:
                    plan_state["tags"] = update_ret["ret"]

        except hub.tool.boto3.exception.ClientError as e:
            result["result"] = False
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)

    else:
        # Policy does not exist , create
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "description": description,
                    "policy_type": policy_type,
                    "content": content,
                    "tags": tags,
                },
            )
            result["comment"] += hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.organizations.policy", name=name
            )
            return result

        create_policy_ret = await hub.exec.boto3.client.organizations.create_policy(
            ctx,
            Name=name,
            Description=description,
            Type=policy_type,
            Content=content,
            Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
            if tags
            else None,
        )
        result["result"] = create_policy_ret["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + create_policy_ret["comment"]
            return result

        resource_id = create_policy_ret["ret"]["Policy"]["PolicySummary"]["Id"]

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.organizations.policy", name=name
        )

    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif (not result["old_state"]) or resource_updated:
        after = await hub.exec.aws.organizations.policy.get(
            ctx=ctx, name=name, resource_id=resource_id
        )
        if not after["result"] or not after["ret"]:
            result["result"] = False
            result["comment"] = after["comment"]
            return result
        result["new_state"] = copy.deepcopy(after["ret"])
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])

    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    """Deletes the specified policy from your organization.

    Before you perform this operation, you must first detach the policy from all organizational units (OUs), roots,
    and accounts. This operation can be called only from the organization's management account.

    Args:
        name(str): The name of the policy.
        resource_id(str): The ID of the policy in Amazon Web Services.

    Request syntax:
      .. code-block:: sls

          [idem_test_aws_organizations_policy]:
            aws.organizations.policy.absent:
              - name: 'string'
              - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem_test_aws_organizations_policy:
              aws.organizations.policy.absent:
                - name: 'idem_test_policy'
                - resource_id: 'p-123456789012'
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.organizations.policy", name=name
        )
        return result

    before = await hub.exec.aws.organizations.policy.get(
        ctx=ctx, name=name, resource_id=resource_id
    )

    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result

    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.organizations.policy", name=name
        )
    else:
        result["old_state"] = copy.deepcopy(before["ret"])
        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                resource_type="aws.organizations.policy", name=name
            )
            return result

        delete_ret = await hub.exec.boto3.client.organizations.delete_policy(
            ctx, PolicyId=resource_id
        )
        if not delete_ret["result"]:
            result["comment"] = result["comment"] + delete_ret["comment"]
            result["result"] = delete_ret["result"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.organizations.policy", name=name
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Gets information about the Organization Policy.

    Describes AWS Organization Policies in a way that can be recreated/managed with the corresponding present
    function.

    Returns:
        Dict[str, Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws.organizations.policy
    """
    result = {}

    list_scp_policies = await hub.exec.boto3.client.organizations.list_policies(
        ctx, Filter="SERVICE_CONTROL_POLICY"
    )

    if not list_scp_policies:
        hub.log.debug(
            f"Could not describe policy with error: {list_scp_policies['comment']}"
        )
        return {}

    list_all_policies = list_scp_policies["ret"]["Policies"]

    list_tag_policies = await hub.exec.boto3.client.organizations.list_policies(
        ctx, Filter="TAG_POLICY"
    )

    if not list_tag_policies:
        hub.log.debug(
            f"Could not describe policy with error: {list_tag_policies['comment']}"
        )
        return {}

    if list_tag_policies["ret"]["Policies"]:
        list_all_policies.extend(list_tag_policies["ret"]["Policies"])

    for policy in list_all_policies:
        policy_id = policy["Id"]

        policy_ret = await hub.exec.boto3.client.organizations.describe_policy(
            ctx, PolicyId=policy_id
        )

        if not policy_ret and not policy_ret["ret"]:
            hub.log.debug(
                f"Could not describe '{policy_id}' with error: {policy_ret['comment']}"
            )
            continue

        tags_ret = await hub.exec.boto3.client.organizations.list_tags_for_resource(
            ctx, ResourceId=policy_id
        )
        if not tags_ret["result"]:
            hub.log.debug(
                f"Unable to list tags for resource {policy_id} with error: {tags_ret['comment']}"
            )

        translated_resource = (
            hub.tool.aws.organizations.conversion_utils.convert_raw_policy_to_present(
                policy_ret["ret"]["Policy"],
                tags_ret["ret"]["Tags"] if tags_ret else None,
            )
        )

        result[policy["Name"]] = {
            "aws.organizations.policy.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in translated_resource.items()
            ]
        }

    return result



"""State module for managing Amazon Elasticache Replication Groups."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]

TREQ = {
    "present": {
        "require": [
            "aws.elasticache.cache_subnet_group.present",
            "aws.elasticache.cache_parameter_group.present",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    replication_group_description: str,
    resource_id: str = None,
    global_replication_group_id: str = None,
    primary_cluster_id: str = None,
    automatic_failover_enabled: bool = None,
    multi_az_enabled: bool = None,
    num_cache_clusters: int = None,
    preferred_cache_cluster_a_zs: List[str] = None,
    num_node_groups: int = None,
    replicas_per_node_group: int = None,
    node_group_configuration: List[
        make_dataclass(
            """A list of node group (shard) configuration options."""
            "NodeGroupConfiguration",
            [
                ("NodeGroupId", str, field(default=None)),
                ("Slots", str, field(default=None)),
                ("ReplicaCount", int, field(default=None)),
                ("PrimaryAvailabilityZone", str, field(default=None)),
                ("ReplicaAvailabilityZones", List[str], field(default=None)),
                ("PrimaryOutpostArn", str, field(default=None)),
                ("ReplicaOutpostArns", List[str], field(default=None)),
            ],
        )
    ] = None,
    cache_node_type: str = None,
    engine: str = None,
    engine_version: str = None,
    cache_parameter_group_name: str = None,
    cache_subnet_group_name: str = None,
    cache_security_group_names: List[str] = None,
    security_group_ids: List[str] = None,
    tags: Dict[str, str] = None,
    snapshot_arns: List[str] = None,
    snapshot_name: str = None,
    preferred_maintenance_window: str = None,
    port: int = None,
    notification_topic_arn: str = None,
    auto_minor_version_upgrade: bool = None,
    snapshot_retention_limit: int = None,
    snapshot_window: str = None,
    auth_token: str = None,
    transit_encryption_enabled: bool = None,
    at_rest_encryption_enabled: bool = None,
    kms_key_id: str = None,
    user_group_ids: List[str] = None,
    log_delivery_configurations: List[
        make_dataclass(
            """Specifies the destination, format and type of the logs."""
            "LogDeliveryConfigurationRequest",
            [
                ("LogType", str, field(default=None)),
                ("DestinationType", str, field(default=None)),
                (
                    "DestinationDetails",
                    make_dataclass(
                        """Configuration details of either a CloudWatch Logs destination or Kinesis Data Firehose destination."""
                        "DestinationDetails",
                        [
                            (
                                "CloudWatchLogsDetails",
                                make_dataclass(
                                    """The configuration details of the CloudWatch Logs destination."""
                                    "CloudWatchLogsDestinationDetails",
                                    [("LogGroup", str, field(default=None))],
                                ),
                                field(default=None),
                            ),
                            (
                                "KinesisFirehoseDetails",
                                make_dataclass(
                                    """The configuration details of the Kinesis Data Firehose destination."""
                                    "KinesisFirehoseDestinationDetails",
                                    [("DeliveryStream", str, field(default=None))],
                                ),
                                field(default=None),
                            ),
                        ],
                    ),
                    field(default=None),
                ),
                ("LogFormat", str, field(default=None)),
                ("Enabled", bool, field(default=None)),
            ],
        )
    ] = None,
    data_tiering_enabled: bool = None,
    snapshotting_cluster_id: str = None,
    node_group_id: str = None,
    notification_topic_status: str = None,
    apply_immediately: bool = None,
    auto_token_update_strategy: str = None,
    user_group_ids_to_add: List[str] = None,
    user_group_ids_to_remove: List[str] = None,
    remove_user_groups: bool = None,
    timeout: make_dataclass(
        "Timeout",
        [
            (
                "create",
                make_dataclass(
                    "CreateTimeout",
                    [
                        ("delay", int, field(default=15)),
                        ("max_attempts", int, field(default=40)),
                    ],
                ),
                field(default=None),
            ),
            (
                "update",
                make_dataclass(
                    "UpdateTimeout",
                    [
                        ("delay", int, field(default=15)),
                        ("max_attempts", int, field(default=40)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
) -> Dict[str, Any]:
    """Creates a Redis (cluster mode disabled) or a Redis (cluster mode enabled) replication group.

    This resource can be used to create a standalone regional replication group or a secondary replication group associated
    with a Global datastore.

    A Redis (cluster mode disabled) replication group is a collection of clusters, where one of the
    clusters is a read/write primary and the others are read-only replicas. Writes to the primary are asynchronously
    propagated to the replicas.

    A Redis cluster-mode enabled cluster is comprised of from 1 to 90 shards. Each shard has a primary node and up to
    5 read-only replica nodes. The configuration can range from 90 shards and 0 replicas to 15 shards and 5 replicas,
    which is the maximum number or replicas allowed

    The node or shard limit can be increased to a maximum of 500 per cluster if the Redis engine version is 5.0.6 or
    higher. For example, you can choose to configure a 500 node cluster that ranges between 83 shards (one primary
    and 5 replicas per shard) and 500 shards (single primary and no replicas). Make sure there are enough available
    IP addresses to accommodate the increase. Common pitfalls include the subnets in the subnet group have too small
    a CIDR range or the subnets are shared and heavily used by other clusters. For more information, see Creating a
    Subnet Group. For versions below 5.0.6, the limit is 250 per cluster.

    To request a limit increase, see Amazon Service Limits and choose the limit type Nodes per cluster per instance type.

    When a Redis (cluster mode disabled) replication group has been successfully created, you can add one or more read replicas to it, up to a
    total of 5 read replicas. If you need to increase or decrease the number of node groups (console: shards), you
    can avail yourself of ElastiCache for Redis' scaling. For more information, see Scaling ElastiCache for Redis
    Clusters in the ElastiCache User Guide.  This operation is valid for Redis only.

    Args:
        name(str):
            An Idem name of the resource. It is used as the Replication Group ID during resource creation.
        resource_id(str, Optional):
            The ID of the replication group in Amazon Web Services.
        replication_group_description(str):
            A user-created description for the replication group.
        global_replication_group_id(str, Optional):
            The name of the Global datastore.
        primary_cluster_id(str, Optional):
            The identifier of the cluster that serves as the primary for this replication group. This
            cluster must already exist and have a status of ``available``.

            This parameter is not required if ``num_cache_clusters``, ``num_node_groups``, or ``replicas_per_node_group`` is specified.
        automatic_failover_enabled(bool, Optional):
            Specifies whether a read-only replica is automatically promoted to read/write primary if the existing primary fails.

            .. warning::
                ``automatic_failover_enabled`` must be enabled for Redis (cluster mode enabled) replication groups.

        multi_az_enabled(bool, Optional):
            A flag indicating if you have Multi-AZ enabled to enhance fault tolerance.
        num_cache_clusters(int, Optional):
            The number of clusters this replication group initially has.

            This parameter is not used if there is more than one node group (shard). You should use ``replicas_per_node_group`` instead.

            If ``automatic_failover_enabled`` is ``True``, the value of this parameter must be at least ``2``.
            If ``automatic_failover_enabled`` is ``True``, you can omit this parameter (it will default to ``1``), or you can
            explicitly set it to a value between ``2`` and ``6``.

            The maximum permitted value for ``num_cache_clusters`` is ``6`` (``1`` primary plus ``5`` replicas).
        preferred_cache_cluster_a_zs(list, Optional):
            A list of EC2 Availability Zones in which the replication group's clusters are created. The
            order of the Availability Zones in the list is the order in which clusters are allocated. The
            primary cluster is created in the first AZ in the list.

            This parameter is not used if there is more than one node group (shard). You should use ``node_group_configuration`` instead.

            .. note::
                If you are creating your replication group in an Amazon VPC (recommended), you can only locate clusters in
                Availability Zones associated with the subnets in the selected subnet group. The number of
                Availability Zones listed must equal the value of ``num_cache_clusters``.
        num_node_groups(int, Optional):
            An optional parameter that specifies the number of node groups (shards) for this Redis (cluster
            mode enabled) replication group. For Redis (cluster mode disabled) either omit this parameter or
            set it to ``1``.
        replicas_per_node_group(int, Optional):
            An optional parameter that specifies the number of replica nodes in each node group (shard).
            Valid values are ``0`` to ``5``.
        node_group_configuration(list, Optional):
            A list of node group (shard) configuration options.

            Each node group (shard) configuration has the following members: ``PrimaryAvailabilityZone``,
            ``ReplicaAvailabilityZones``, ``ReplicaCount``, and ``Slots``.

            If you're creating a Redis (cluster mode disabled) or a Redis (cluster mode enabled) replication group,
            you can use this parameter to individually configure each node group (shard), or you can omit this parameter.
            However, it is required when seeding a Redis (cluster mode enabled) cluster from a S3 rdb file. You must
            configure each node group (shard) using this parameter because you must specify the slots for each node group.

            * NodeGroupId (*str, Optional*):
                Either the ElastiCache for Redis supplied 4-digit id or a user supplied id for the node group these configuration values apply to.
            * Slots (*str, Optional*):
                A string that specifies the keyspace for a particular node group. Keyspaces range from ``0`` to ``16,383``.
                The string is in the format ``startkey-endkey``.
            * ReplicaCount (*int, Optional*):
                The number of read replica nodes in this node group (shard).
            * PrimaryAvailabilityZone (*str, Optional*):
                The Availability Zone where the primary node of this node group (shard) is launched.
            * ReplicaAvailabilityZones (*list, Optional*):
                A list of Availability Zones to be used for the read replicas. The number of Availability Zones in this list
                must match the value of ``ReplicaCount`` or ``ReplicasPerNodeGroup`` if not specified.
            * PrimaryOutpostArn (*str, Optional*):
                The outpost ARN of the primary node.
            * ReplicaOutpostArns (*list, optional*):
                The outpost ARN of the node replicas.

        cache_node_type(str, Optional):
            The compute and memory capacity of the nodes in the node group (shard).

            The following node types are supported by ElastiCache. Generally speaking, the current generation types provide more
            memory and computational power at lower cost when compared to their equivalent previous generation counterparts.

            * General purpose:
                - Current generation:
                    * **M6g node types** (available only for Redis engine version 5.0.6 onward and for Memcached engine version 1.5.16 onward): ``cache.m6g.large``,
                        ``cache.m6g.xlarge``, ``cache.m6g.2xlarge``, ``cache.m6g.4xlarge``, ``cache.m6g.8xlarge``, ``cache.m6g.12xlarge``, ``cache.m6g.16xlarge``.
                    * **M5 node types**: ``cache.m5.large``, ``cache.m5.xlarge``, ``cache.m5.2xlarge``, ``cache.m5.4xlarge``, ``cache.m5.12xlarge``, ``cache.m5.24xlarge``.
                    * **M4 node types**: ``cache.m4.large``, ``cache.m4.xlarge``, ``cache.m4.2xlarge``, ``cache.m4.4xlarge``, ``cache.m4.10xlarge``.
                    * **T4g node types** (available only for Redis engine version 5.0.6 onward and Memcached engine version 1.5.16 onward): ``cache.t4g.micro``,
                        ``cache.t4g.small``, ``cache.t4g.medium``.
                    * **T3 node types**: ``cache.t3.micro``, ``cache.t3.small``, ``cache.t3.medium``.
                    * **T2 node types**: ``cache.t2.micro``, ``cache.t2.small``, ``cache.t2.medium``.
                - Previous generation: (not recommended. Existing clusters are still supported but creation of new clusters is not supported for these types.):
                    * **T1 node types**: ``cache.t1.micro``.
                    * **M1 node types**: ``cache.m1.small``, ``cache.m1.medium``, ``cache.m1.large``, ``cache.m1.xlarge``.
                    * **M3 node types**: ``cache.m3.medium``, ``cache.m3.large``, ``cache.m3.xlarge``, ``cache.m3.2xlarge``.
            * Compute optimized:
                - Previous generation: (not recommended. Existing clusters are still supported but creation of new clusters is not supported for these types.)
                    * **C1 node types**: ``cache.c1.xlarge``
            * Memory optimized with data tiering:
                - Current generation:
                    * **R6gd node types** (available only for Redis engine version 6.2 onward:) ``cache.r6gd.xlarge``, ``cache.r6gd.2xlarge``,
                        ``cache.r6gd.4xlarge``, ``cache.r6gd.8xlarge``, ``cache.r6gd.12xlarge``, ``cache.r6gd.16xlarge``.
            * Memory optimized:
                - Current generation:
                    * **R6g node types** (available only for Redis engine version 5.0.6 onward and for Memcached engine version 1.5.16 onward): ``cache.r6g.large``,
                        ``cache.r6g.xlarge``, ``cache.r6g.2xlarge``, ``cache.r6g.4xlarge``, ``cache.r6g.8xlarge``, ``cache.r6g.12xlarge``, ``cache.r6g.16xlarge``.
                    * **R5 node types**: ``cache.r5.large``, ``cache.r5.xlarge``, ``cache.r5.2xlarge``, ``cache.r5.4xlarge``, ``cache.r5.12xlarge``, ``cache.r5.24xlarge``.
                    * **R4 node types**: ``cache.r4.large``, ``cache.r4.xlarge``, ``cache.r4.2xlarge``, ``cache.r4.4xlarge``, ``cache.r4.8xlarge``, ``cache.r4.16xlarge``.
                - Previous generation (not recommended. Existing clusters are still supported but creation of new clusters is not supported for these types.):
                    * **M2 node types**: ``cache.m2.xlarge``, ``cache.m2.2xlarge``, ``cache.m2.4xlarge``.
                    * **R3 node types**: ``cache.r3.large``, ``cache.r3.xlarge``, ``cache.r3.2xlarge``, ``cache.r3.4xlarge``, ``cache.r3.8xlarge``.

            .. note::
                Additional node type info

                * All current generation instance types are created in Amazon VPC by default.
                * Redis append-only files (AOF) are not supported for T1 or T2 instances.
                * Redis Multi-AZ with automatic failover is not supported on T1 instances.
                * Redis configuration variables ``appendonly`` and ``appendfsync`` are not supported on Redis version 2.8.22 and later.

        engine(str, Optional):
            The name of the cache engine to be used for the clusters in this replication group. Must be ``Redis``.
        engine_version(str, Optional):
            The version number of the cache engine to be used for the clusters in this replication group.

            .. warning::
                You can upgrade to a newer engine version (see Selecting a Cache Engine and Version
                in the ElastiCache User Guide), but you cannot downgrade to an earlier engine version. If you
                want to use an earlier engine version, you must delete the existing cluster or replication group
                and create it anew with the earlier engine version.
        cache_parameter_group_name(str, Optional):
            The name of the parameter group to associate with this replication group. If this argument is
            omitted, the default cache parameter group for the specified engine is used. If you are running
            Redis version 3.2.4 or later, only one node group (shard), and want to use a default parameter
            group, we recommend that you specify the parameter group by name.

            * To create a Redis (cluster mode disabled) replication group, use ``cache_parameter_group_name=default.redis3.2``.
            * To create a Redis (cluster mode enabled) replication group, use ``cache_parameter_group_name=default.redis3.2.cluster.on``.

        cache_subnet_group_name(str, Optional):
            The name of the cache subnet group to be used for the replication group.

            .. warning::
                If you're going to launch your cluster in an Amazon VPC, you need to create a subnet group before you start
                creating a cluster. For more information, see Subnets and Subnet Groups.

        cache_security_group_names(list, Optional):
            A list of cache security group names to associate with this replication group.
        security_group_ids(list, Optional):
            One or more Amazon VPC security groups associated with this replication group. Use this
            parameter only when you are creating a replication group in an Amazon Virtual Private Cloud
            (Amazon VPC).
        tags(dict or list, Optional):
            Dict in the format of ``{tag-key: tag-value}`` or List of tags in the format of
            ``[{"Key": tag-key, "Value": tag-value}]`` to associate with the replication group.

            * Key (*str*):
                The key of the tag.
            * Value (*str*):
                The value of the tag.

        snapshot_arns(list, Optional):
            A list of Amazon Resource Names (ARN) that uniquely identify the Redis RDB snapshot files stored
            in Amazon S3. The snapshot files are used to populate the new replication group. The Amazon S3
            object name in the ARN cannot contain any commas. The new replication group will have the number
            of node groups specified by the parameter ``num_node_groups`` or the number of node
            groups configured by ``node_group_configuration`` regardless of the number of ARNs specified here.

            Example of an Amazon S3 ARN: ``arn:aws:s3:::my_bucket/snapshot1.rdb``.
        snapshot_name(str, Optional):
            The name of a snapshot from which to restore data into the new replication group. The snapshot
            status changes to restoring while the new replication group is being created.
        preferred_maintenance_window(str, Optional):
            Specifies the weekly time range during which maintenance on the cluster is performed. It is
            specified as a range in the format ``ddd:hh24:mi-ddd:hh24:mi`` (24H Clock UTC). The minimum
            maintenance window is a 60 minute period. Valid values for ``ddd`` are: ``sun``, ``mon``, ``tue``, ``wed``, ``thu``, ``fri``, ``sat``.

            Example: ``sun:23:00-mon:01:30``.
        port(int, Optional):
            The port number on which each member of the replication group accepts connections.
        notification_topic_arn(str, Optional):
            The Amazon Resource Name (ARN) of the Amazon Simple Notification Service (SNS) topic to which
            notifications are sent.

            .. note::
                The Amazon SNS topic owner must be the same as the cluster owner.

        auto_minor_version_upgrade(bool, Optional):
            If you are running Redis engine version 6.0 or later, set this parameter to yes if you want to
            opt-in to the next auto minor version upgrade campaign. This parameter is disabled for previous
            versions.
        snapshot_retention_limit(int, Optional):
            The number of days for which ElastiCache retains automatic snapshots before deleting them. For
            example, if you set ``snapshot_retention_limit`` to ``5``, a snapshot that was taken today is retained for
            5 days before being deleted.

            Default: ``0`` (i.e., automatic backups are disabled for this cluster).
        snapshot_window(str, Optional):
            The daily time range (in UTC) during which ElastiCache begins taking a daily snapshot of your
            node group (shard).
            Example: ``05:00-09:00``.
            If you do not specify this parameter, ElastiCache automatically chooses an appropriate time range.
        auth_token(str, Optional):
            .. note::
                Reserved parameter. The password used to access a password protected server. ``auth_token`` can be
                specified only on replication groups where ``transit_encryption_enabled`` is ``True``.

            Password constraints:
                * Must be only printable ASCII characters.
                * Must be at least 16 characters and no more than 128 characters in length.
                * The only permitted printable special characters are ``!``, ``&``, ``#``, ``$``, ``^``, ``<``, ``>``, and ``-``. Other printable special characters cannot be  used in the AUTH token.

            For more information, see AUTH password at http://redis.io/commands/AUTH.
        transit_encryption_enabled(bool, Optional):
            A flag that enables in-transit encryption when set to true.

            You cannot modify the value of ``transit_encryption_enabled`` after the cluster is created. To enable in-transit encryption on a
            cluster you must set ``transit_encryption_enabled`` to ``True`` when you create a cluster.

            This parameter is valid only if the ``engine`` parameter is ``redis``, the ``engine_version`` parameter is ``3.2.6``, ``4.x`` or
            later, and the cluster is being created in an Amazon VPC. If you enable in-transit encryption,
            you must also specify a value for ``cache_subnet_group_name``.

            .. note::
                Required: Only available when creating a replication group in an Amazon VPC using redis version ``3.2.6``, ``4.x`` or later.

            .. warning::
                For HIPAA compliance, you must specify ``transit_encryption_enabled`` as ``True``, an ``auth_token``, and a ``cache_subnet_group_name``.

        at_rest_encryption_enabled(bool, Optional):
            A flag that enables encryption at rest when set to ``True``.

            You cannot modify the value of ``at_rest_encryption_enabled`` after the replication group is created.
            To enable encryption at rest on a replication group you must set ``at_rest_encryption_enabled`` to ``True`` when you
            create the replication group.

            .. note::
                Required: Only available when creating a replication group in an Amazon VPC using redis version ``3.2.6``, ``4.x`` or later.
        kms_key_id(str, Optional):
            The ID of the KMS key used to encrypt the disk in the cluster.
        user_group_ids(list, Optional):
            The user group to associate with the replication group.
        log_delivery_configurations(list, Optional):
            Specifies the destination, format and type of the logs.

            * LogType (*str, Optional*):
                Refers to slow-log.
            * DestinationType (*str, Optional*):
                Specify either cloudwatch-logs or kinesis-firehose as the destination type.
            * DestinationDetails (*dict, Optional*):
                Configuration details of either a CloudWatch Logs destination or Kinesis Data Firehose destination.

                * CloudWatchLogsDetails (*dict, Optional*):
                    The configuration details of the CloudWatch Logs destination.

                    * LogGroup (*str, Optional*):
                        The name of the CloudWatch Logs log group.
                * KinesisFirehoseDetails (*dict, Optional*):
                    The configuration details of the Kinesis Data Firehose destination.

                    * DeliveryStream (*str, Optional*):
                        The name of the Kinesis Data Firehose delivery stream.

            * LogFormat (*str, Optional*):
                Specifies either ``JSON`` or ``TEXT``.
            * Enabled (*bool, Optional*):
                Specify if log delivery is enabled.

        data_tiering_enabled(bool, Optional):
            Enables data tiering. Data tiering is only supported for replication groups using the ``r6gd`` node
            type. This parameter must be set to ``True`` when using ``r6gd`` nodes. For more information, see Data
            tiering.
        snapshotting_cluster_id(str, Optional):
            The cluster ID that is used as the daily snapshot source for the replication group. This parameter
            cannot be set for Redis (cluster mode enabled) replication groups.
        node_group_id(str, Optional):
            Deprecated. This parameter is not used.
        notification_topic_status(bool, Optional):
            The Amazon Resource Name (ARN) of the Amazon SNS topic to which notifications are sent.
        apply_immediately(bool, Optional):
            If ``True``, this parameter causes the modifications in this request and any pending modifications to be applied,
            asynchronously and as soon as possible, regardless of the ``preferred_maintenance_window`` setting for the replication group.
            If ``False``, changes to the nodes in the replication group are applied on the next maintenance reboot, or the next failure reboot,
            whichever occurs first.
        auto_token_update_strategy(str, Optional):
            Specifies the strategy to use to update the AUTH token. This parameter must be specified with
            the ``auth_token`` parameter. Possible values: ``Rotate``, and ``Set``. For more information, see Authenticating Users with Redis AUTH
        user_group_ids_to_add(list, Optional):
            The ID of the user group you are associating with the replication group.
        user_group_ids_to_remove(list, Optional):
            The ID of the user group to disassociate from the replication group, meaning the users in the group no longer can
            access the replication group.
        remove_user_groups(bool, Optional):
            Removes the user group associated with this replication group.
        timeout(dict, Optional):
            Timeout configuration for AWS Elasticache Replication Group operations.

            * create (*dict, Optional*):
                Timeout configuration when creating an AWS Elasticache Replication Group.

                * delay (*int, Optional*):
                    The amount of time in seconds to wait between attempts. Default value is ``15``.
                * max_attempts (*int, Optional*):
                    Max attempts of waiting for change. Default value is ``40``.

            * update (*dict, Optional*):
                Timeout configuration when updating an AWS Elasticache Replication Group.

                * delay (*int, Optional*):
                    The amount of time in seconds to wait between attempts. Default value is ``15``.
                * max_attempts (*int, Optional*):
                    Max attempts of waiting for change. Default value is ``40``.

    Request Syntax:
      .. code-block:: sls

        [idem_test_aws_elasticache_replication_group]:
          aws.elasticache.replication_group.present:
            - name: 'string'
            - resource_id: 'string'
            - replication_group_description: 'string'
            - global_replication_group_id: 'string'
            - primary_cluster_id: 'string'
            - automatic_failover_enabled: True|False
            - multi_az_enabled: True|False
            - num_cache_clusters: int
            - preferred_cache_cluster_a_zs:
                - 'string'
            - num_node_groups: int
            - replicas_per_node_group: int
            - node_group_configuration:
                - NodeGroupId: 'string'
                  Slots: 'string'
                  ReplicaCount: 'int'
                  PrimaryAvailabilityZone: 'string'
                  ReplicaAvailabilityZones:
                    - 'string'
                  PrimaryOutpostArn: 'string'
                  ReplicaOutpostArns:
                    - 'string'
            - cache_node_type: 'string'
            - engine: 'string'
            - engine_version: 'string'
            - cache_parameter_group_name: 'string'
            - cache_subnet_group_name: 'string'
            - cache_security_group_names:
                - 'string'
            - security_group_ids:
                - 'string'
            - tags:
                - Key: 'string'
                  Value: 'string'
            - snapshot_arns:
                - 'string'
            - snapshot_name: 'string'
            - preferred_maintenance_window: 'string'
            - port: int
            - notification_topic_arn: 'string'
            - auto_minor_version_upgrade: True|False
            - snapshot_retention_limit: int
            - snapshot_window: 'string'
            - auth_token: 'string'
            - transit_encryption_enabled: True|False
            - at_rest_encryption_enabled: True|False
            - kms_key_id: 'string'
            - user_group_ids:
                - 'string'
            - log_delivery_configurations:
                - LogType: 'string'
                  DestinationType: 'string'
                  DestinationDetails:
                    CloudWatchLogsDetails:
                      LogGroup: 'string'
                    KinesisFirehoseDetails:
                      DeliveryStream: 'string'
                  LogFormat: 'string'
                  Enabled: True|False
            - data_tiering_enabled: True|False
            - snapshotting_cluster_id: 'string'
            - node_group_id: 'string'
            - notification_topic_status: 'string'
            - apply_immediately: True|False
            - auto_token_update_strategy: 'string'
            - user_group_ids_to_add:
                - 'string'
            - user_group_ids_to_remove:
                - 'string'
            - remove_user_groups: True|False
            - timeout:
                create:
                  delay: int
                  max_attemps: int
                update:
                  delay: int
                  max_attemps: int

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem_test_aws_elasticache_replication_group:
              aws.elasticache.replication_group.present:
                - name: 'idem_test_replication_group'
                - replication_group_description: 'My Elasticache Replication Group'
                - num_node_groups: 1
                - num_cache_clusters: 123
                - automatic_failover_enabled: True
                - multi_az_enabled: True
                - engine: 'Redis'
                - cache_node_type: 'cache.t2.micro'
                - cache_subnet_group_name: 'idem_test_cache_subnet_group'
                - cache_parameter_group_name: 'idem_test_cache_parameter_group'
                - cache_security_group_names: 'idem_test_cache_security_group'
                - tags:
                    - Key: 'provider'
                      Value: 'idem'
    """
    result = dict(comment=[], name=name, old_state=None, new_state=None, result=True)
    before = None
    resource_updated = False
    if resource_id:
        before = await hub.exec.aws.elasticache.replication_group.get(
            ctx, name=name, resource_id=resource_id
        )
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result
        result["old_state"] = copy.deepcopy(before["ret"])
        plan_state = copy.deepcopy(result["old_state"])
        update_ret = (
            await hub.tool.aws.elasticache.elasticache_utils.update_replication_group(
                ctx,
                name=name,
                resource_id=resource_id,
                plan_state=plan_state,
                old_state=result["old_state"],
                replication_group_description=replication_group_description,
                primary_cluster_id=primary_cluster_id,
                snapshotting_cluster_id=snapshotting_cluster_id,
                automatic_failover_enabled=automatic_failover_enabled,
                multi_az_enabled=multi_az_enabled,
                node_group_id=node_group_id,
                cache_security_group_names=cache_security_group_names,
                security_group_ids=security_group_ids,
                preferred_maintenance_window=preferred_maintenance_window,
                notification_topic_arn=notification_topic_arn,
                cache_parameter_group_name=cache_parameter_group_name,
                notification_topic_status=notification_topic_status,
                apply_immediately=apply_immediately,
                engine_version=engine_version,
                auto_minor_version_upgrade=auto_minor_version_upgrade,
                snapshot_retention_limit=snapshot_retention_limit,
                snapshot_window=snapshot_window,
                cache_node_type=cache_node_type,
                auth_token=auth_token,
                auto_token_update_strategy=auto_token_update_strategy,
                user_group_ids_to_add=user_group_ids_to_add,
                user_group_ids_to_remove=user_group_ids_to_remove,
                remove_user_groups=remove_user_groups,
                log_delivery_configurations=log_delivery_configurations,
            )
        )
        resource_updated = resource_updated or update_ret["result"]
        if not update_ret["result"]:
            result["comment"] += list(update_ret["comment"])
            result["result"] = False
            return result
        # Waiting for the ReplicationGroup status to be available after updates
        waiter_ret = (
            await hub.tool.aws.elasticache.elasticache_utils.replication_group_waiter(
                ctx, name, resource_id, timeout, "update"
            )
        )
        if not waiter_ret["result"]:
            result["result"] = False
            result["comment"] += list(waiter_ret["comment"])
        if tags is not None and tags != result["old_state"].get("tags"):
            # Update tags
            update_ret = await hub.tool.aws.elasticache.elasticache_utils.update_tags(
                ctx=ctx,
                resource_arn=result["old_state"].get("arn"),
                old_tags=result["old_state"].get("tags"),
                new_tags=tags,
                waiter=hub.tool.aws.elasticache.elasticache_utils.replication_group_waiter,
                waiter_args={
                    "ctx": ctx,
                    "name": name,
                    "resource_id": resource_id,
                    "timeout": timeout,
                    "operation_type": "update",
                },
            )
            result["comment"] += list(update_ret["comment"])
            result["result"] = update_ret["result"]
            resource_updated = bool(update_ret["result"])
            if ctx.get("test", False) and resource_updated:
                plan_state["tags"] = update_ret["ret"]
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "replication_group_id": name,
                    "replication_group_description": replication_group_description,
                    "engine": engine,
                    "cache_node_type": cache_node_type,
                    "cache_subnet_group_name": cache_subnet_group_name,
                    "security_group_ids": security_group_ids,
                    "global_replication_group_id": global_replication_group_id,
                    "primary_cluster_id": primary_cluster_id,
                    "automatic_failover_enabled": automatic_failover_enabled,
                    "num_cache_clusters": num_cache_clusters,
                    "preferred_cache_cluster_a_zs": preferred_cache_cluster_a_zs,
                    "num_node_groups": num_node_groups,
                    "replicas_per_node_group": replicas_per_node_group,
                    "node_group_configuration": node_group_configuration,
                    "engine_version": engine_version,
                    "cache_parameter_group_name": cache_parameter_group_name,
                    "cache_security_group_names": cache_security_group_names,
                    "tags": tags,
                    "snapshot_arns": snapshot_arns,
                    "snapshot_name": snapshot_name,
                    "preferred_maintenance_window": preferred_maintenance_window,
                    "port": port,
                    "notification_topic_arn": notification_topic_arn,
                    "auto_minor_version_upgrade": auto_minor_version_upgrade,
                    "snapshot_retention_limit": snapshot_retention_limit,
                    "snapshot_window": snapshot_window,
                    "auth_token": auth_token,
                    "transit_encryption_enabled": transit_encryption_enabled,
                    "at_rest_encryption_enabled": at_rest_encryption_enabled,
                    "kms_key_id": kms_key_id,
                    "user_group_ids": user_group_ids,
                    "log_delivery_configurations": log_delivery_configurations,
                    "data_tiering_enabled": data_tiering_enabled,
                },
            )
            result["comment"] = list(
                hub.tool.aws.comment_utils.would_create_comment(
                    resource_type="aws.elasticache.replication_group", name=name
                )
            )
            return result
        ret = await hub.exec.boto3.client.elasticache.create_replication_group(
            ctx,
            ReplicationGroupId=name,
            ReplicationGroupDescription=replication_group_description,
            Engine=engine,
            CacheNodeType=cache_node_type,
            CacheSubnetGroupName=cache_subnet_group_name,
            SecurityGroupIds=security_group_ids,
            GlobalReplicationGroupId=global_replication_group_id,
            PrimaryClusterId=primary_cluster_id,
            AutomaticFailoverEnabled=automatic_failover_enabled,
            MultiAZEnabled=multi_az_enabled,
            NumCacheClusters=num_cache_clusters,
            PreferredCacheClusterAZs=preferred_cache_cluster_a_zs,
            NumNodeGroups=num_node_groups,
            ReplicasPerNodeGroup=replicas_per_node_group,
            NodeGroupConfiguration=node_group_configuration,
            EngineVersion=engine_version,
            CacheParameterGroupName=cache_parameter_group_name,
            CacheSecurityGroupNames=cache_security_group_names,
            Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags),
            SnapshotArns=snapshot_arns,
            SnapshotName=snapshot_name,
            PreferredMaintenanceWindow=preferred_maintenance_window,
            Port=port,
            NotificationTopicArn=notification_topic_arn,
            AutoMinorVersionUpgrade=auto_minor_version_upgrade,
            SnapshotRetentionLimit=snapshot_retention_limit,
            SnapshotWindow=snapshot_window,
            AuthToken=auth_token,
            TransitEncryptionEnabled=transit_encryption_enabled,
            AtRestEncryptionEnabled=at_rest_encryption_enabled,
            KmsKeyId=kms_key_id,
            UserGroupIds=user_group_ids,
            LogDeliveryConfigurations=log_delivery_configurations,
            DataTieringEnabled=data_tiering_enabled,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = list(ret["comment"])
            return result
        resource_id = name
        # This makes sure the created replication group is saved to esm regardless if the subsequent update call
        # fails or not.
        result["new_state"] = {"name": name, "resource_id": resource_id}
        # Waiting until the ReplicationGroup creation
        waiter_ret = (
            await hub.tool.aws.elasticache.elasticache_utils.replication_group_waiter(
                ctx, name, resource_id, timeout, "create"
            )
        )
        if not waiter_ret["result"]:
            result["result"] = False
            result["comment"] = list(waiter_ret["comment"])
        result["comment"] = list(
            hub.tool.aws.comment_utils.create_comment(
                resource_type="aws.elasticache.replication_group", name=name
            )
        )
    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif (not before) or resource_updated:
        after = await hub.exec.aws.elasticache.replication_group.get(
            ctx, name=name, resource_id=resource_id
        )
        if not after["result"]:
            result["result"] = False
            result["comment"] += list(after["comment"])
            return result
        result["new_state"] = copy.deepcopy(after["ret"])
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    retain_primary_cluster: bool = None,
    final_snapshot_identifier: str = None,
    timeout: make_dataclass(
        """Timeout configuration for AWS Elasticache Replication Group operations."""
        "Timeout",
        [
            (
                "delete",
                make_dataclass(
                    """Timeout configuration when deleting an AWS Elasticache Replication Group."""
                    "DeleteTimeout",
                    [
                        ("delay", int, field(default=15)),
                        ("max_attempts", int, field(default=40)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
) -> Dict[str, Any]:
    """Deletes the specified AWS Elasticache Replication Group.

    By default, this operation deletes the entire replication group, including the primary/primaries and all of the read replicas.
    If the replication group has only one primary, you can optionally delete only the read replicas, while retaining the primary
    by setting ``retain_primary_cluster: True``.
    When you receive a successful response from this operation, Amazon ElastiCache immediately begins deleting the selected resources;
    you cannot cancel or revert this operation. This operation is valid for Redis only.

    Args:
        name(str):
            An Idem name of the resource.
        resource_id(str, Optional):
            The ID of the replication group in Amazon Web Services.

            .. warning::
              Idem automatically considers this resource being absent if this field is not specified.

        retain_primary_cluster(bool, Optional):
            If set to ``True``, all of the read replicas are deleted, but the primary node is retained.
        final_snapshot_identifier(str, Optional):
            The name of a final node group (shard) snapshot. ElastiCache creates the snapshot from the
            primary node in the cluster, rather than one of the replicas; this is to ensure that it captures
            the freshest data. After the final snapshot is taken, the replication group is immediately
            deleted.
        timeout(dict, Optional):
            Timeout configuration for AWS Elasticache Replication Group operations.

            * delete (*dict, Optional*):
                Timeout configuration when deleting an AWS Elasticache Replication Group.

                * delay (*int, Optional*) -- The amount of time in seconds to wait between attempts. Default value is ``15``.
                * max_attempts (*int, Optional*) -- Max attempts of waiting for change. Default value is ``40``.

    Request Syntax:
      .. code-block:: sls

        [idem_test_aws_elasticache_replication_group]:
          aws.elasticache.replication_group.absent:
            - name: 'string'
            - resource_id: 'string'
            - retain_primary_cluster: True|False
            - final_snapshot_identifier: 'string'
            - timeout:
                delete:
                  delay: int
                  max_attemps: int

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem_test_aws_elasticache_replication_group:
              aws.elasticache.replication_group.absent:
                - name: 'idem_test_replication_group'
                - resource_id: 'idem_test_replication_group'
    """
    result = dict(comment=[], name=name, old_state=None, new_state=None, result=True)

    if not resource_id:
        result["comment"] = list(
            hub.tool.aws.comment_utils.already_absent_comment(
                resource_type="aws.elasticache.replication_group", name=name
            )
        )
        return result
    before = await hub.exec.aws.elasticache.replication_group.get(
        ctx, name=name, resource_id=resource_id
    )
    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result

    if not before["ret"]:
        result["comment"] = list(
            hub.tool.aws.comment_utils.already_absent_comment(
                resource_type="aws.elasticache.replication_group", name=name
            )
        )
    elif ctx.get("test", False):
        result["old_state"] = before["ret"]
        result["comment"] = list(
            hub.tool.aws.comment_utils.would_delete_comment(
                resource_type="aws.elasticache.replication_group", name=name
            )
        )
        return result
    else:
        result["old_state"] = before["ret"]
        ret = await hub.exec.boto3.client.elasticache.delete_replication_group(
            ctx,
            ReplicationGroupId=resource_id,
            RetainPrimaryCluster=retain_primary_cluster,
            FinalSnapshotIdentifier=final_snapshot_identifier,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] += list(ret["comment"])
            result["result"] = False
            return result

        # Waiting for the ReplicationGroup to Delete
        waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
            default_delay=15,
            default_max_attempts=40,
            timeout_config=timeout.get("delete") if timeout else None,
        )
        try:
            await hub.tool.boto3.client.wait(
                ctx,
                "elasticache",
                "replication_group_deleted",
                ReplicationGroupId=resource_id,
                WaiterConfig=waiter_config,
            )
        except Exception as e:
            result["comment"] += list(str(e))
            result["result"] = False
        result["comment"] = list(
            hub.tool.aws.comment_utils.delete_comment(
                resource_type="aws.elasticache.replication_group", name=name
            )
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describes Elasticache Replication Groups in a way that can be recreated/managed with the corresponding "present" function.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe aws.elasticache.replication_group
    """
    result = {}
    ret = await hub.exec.boto3.client.elasticache.describe_replication_groups(ctx)

    if not ret["result"]:
        hub.log.debug(
            f"Could not describe aws.elasticache.replication_group {ret['comment']}"
        )
        return {}

    for resource in ret["ret"]["ReplicationGroups"]:
        resource_id = resource.get("ReplicationGroupId")
        tag_ret = await hub.exec.boto3.client.elasticache.list_tags_for_resource(
            ctx, ResourceName=resource.get("ARN")
        )
        if not tag_ret["result"]:
            # Error fetching tags. If fetching tags itself fails, just skip this and continue. Since tags are
            # optional, they need not be associated with all replication groups. But API needs to succeed.
            hub.log.warning(
                f"Failed listing tags for aws.elasticache.replication_group '{resource.get('ARN')}' with error {tag_ret['comment']}"
                f"Describe will skip this aws.elasticache.replication_group and continue. "
            )
            continue
        resource["Tags"] = tag_ret["ret"].get("TagList", [])
        resource_translated = hub.tool.aws.elasticache.conversion_utils.convert_raw_elasticache_replication_group_to_present(
            ctx=ctx,
            raw_resource=resource,
            idem_resource_name=resource_id,
        )
        result[resource_id] = {
            "aws.elasticache.replication_group.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.elbv2.create_load_balancer
hub.exec.boto3.client.elbv2.delete_load_balancer
hub.exec.boto3.client.elbv2.describe_load_balancers
hub.exec.boto3.client.elbv2.describe_load_balancer_attributes
hub.exec.boto3.client.elbv2.modify_load_balancer_attributes
hub.exec.boto3.client.elbv2.set_subnets
hub.exec.boto3.client.elbv2.set_security_groups
hub.exec.boto3.client.elbv2.set_ip_address_type
hub.exec.boto3.client.elbv2.remove_tags
hub.exec.boto3.client.elbv2.describe_tags
hub.exec.boto3.client.elbv2.add_tags


"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    subnets: List[str] = None,
    subnet_mappings: List[
        make_dataclass(
            "SubnetMapping",
            [
                ("SubnetId", str, field(default=None)),
                ("AllocationId", str, field(default=None)),
                ("PrivateIPv4Address", str, field(default=None)),
                ("IPv6Address", str, field(default=None)),
            ],
        )
    ] = None,
    security_groups: List[str] = None,
    scheme: str = None,
    tags: Dict[str, str] = None,
    lb_type: str = None,
    ip_address_type: str = None,
    customer_owned_ipv4_pool: str = None,
    attributes: List[
        make_dataclass(
            "Attribute",
            [
                ("Key", str, field(default=None)),
                ("Value", str, field(default=None)),
            ],
        )
    ] = None,
) -> Dict[str, Any]:
    r"""
    Creates an Application Load Balancer, Network Load Balancer, or Gateway Load Balancer.

    This operation is idempotent, which means that it completes at most one time. If you attempt to create multiple
    load balancers with the same settings, each call succeeds.

    Args:
        name(Text): An Idem name to identify load balancer. This name must be unique per region per account, can have a
            maximum of 32 characters, must contain only alphanumeric characters or hyphens, must not begin or end with
            a hyphen, and must not begin with "internal-"
        subnets(List, Optional): The IDs of the public subnets. You can specify only one subnet per Availability Zone.
            You must specify either subnets or subnet mappings, but not both. To specify an Elastic IP address, specify
            subnet mappings instead of subnets.

            [Application Load Balancers] You must specify subnets from at least two Availability Zones.
            [Application Load Balancers on Outposts] You must specify one Outpost subnet.
            [Application Load Balancers on Local Zones] You can specify subnets from one or more Local Zones.
            [Network Load Balancers] You can specify subnets from one or more Availability Zones.
            [Gateway Load Balancers] You can specify subnets from one or more Availability Zones.
        subnet_mappings([List[Dict[str, Any]]): The IDs of the public subnets. You can specify only one subnet per
            Availability Zone. You must specify either subnets or subnet mappings, but not both.

            [Application Load Balancers] You must specify subnets from at least two Availability Zones. You cannot
                specify Elastic IP addresses for your subnets.
            [Application Load Balancers on Outposts] You must specify one Outpost subnet.
            [Application Load Balancers on Local Zones] You can specify subnets from one or more Local Zones.
            [Network Load Balancers] You can specify subnets from one or more Availability Zones. You can specify one
                Elastic IP address per subnet if you need static IP addresses for your internet-facing load balancer.
                For internal load balancers, you can specify one private IP address per subnet from the IPv4 range of
                the subnet. For internet-facing load balancer, you can specify one IPv6 address per subnet.
            [Gateway Load Balancers] You can specify subnets from one or more Availability Zones. You cannot specify
                Elastic IP addresses for your subnets.

                * SubnetId (Text, Optional): The ID of the subnet.
                * AllocationId (Text, Optional): [Network Load Balancers] The allocation ID of the Elastic IP address
                    for an internet-facing load balancer.
                * PrivateIPv4Address (Text, Optional): [Network Load Balancers] The private IPv4 address for an internal
                    load balancer.
                * IPv6Address (Text, Optional): [Network Load Balancers] The IPv6 address.
        security_groups(List): [Application Load Balancers] The IDs of the security groups for the load balancer.
        scheme(Text, Optional): The nodes of an Internet-facing load balancer have public IP addresses. The DNS name of
            an Internet-facing load balancer is publicly resolvable to the public IP addresses of the nodes. Therefore,
            Internet-facing load balancers can route requests from clients over the internet.

            The nodes of an internal load balancer have only private IP addresses. The DNS name of an internal load
            balancer is publicly resolvable to the private IP addresses of the nodes. Therefore, internal load balancers
             can route requests only from clients with access to the VPC for the load balancer.

            The default is an Internet-facing load balancer. You cannot specify a scheme for a Gateway Load Balancer.
        tags([Dict[str, str], optional): The tags to assign to the load balancer. Defaults to None.
            * Key (Text): The key of the tag.
            * Value (Text, optional): The value of the tag.
        lb_type(Text, Optional): The type of load balancer. The default is application.
        ip_address_type(Text, Optional): The type of IP addresses used by the subnets for your load balancer. The
            possible values are ipv4 (for IPv4 addresses) and dualstack (for IPv4 and IPv6 addresses).
        customer_owned_ipv4_pool(Text, Optional): [Application Load Balancers on Outposts] The ID of the customer-owned
            address pool (CoIP pool).
        attributes: The load balancer attributes.
            (Dict): Information about a tag.
                Key (Text): The name of the attribute.

            Attributes supported by all load balancers.
                1. 'deletion_protection.enabled' - Indicates whether deletion protection is enabled. The value is true
                    or false. The default is false.

            The following attributes are supported by both Application Load Balancers and Network Load Balancers:
                1. 'access_logs.s3.enabled' - Indicates whether access logs are enabled. The value is true or false.
                   The default is false .
                2. 'access_logs.s3.bucket' - The name of the S3 bucket for the access logs. This attribute is required
                   if access logs are enabled. The bucket must exist in the same region as the load balancer and have a
                   bucket policy that grants Elastic Load Balancing permissions to write to the bucket.
                3. 'access_logs.s3.prefix' - The prefix for the location in the S3 bucket for the access logs.
                4. 'ipv6.deny_all_igw_traffic' - Blocks internet gateway (IGW) access to the load balancer. It is set
                   to false for internet-facing load balancers and true for internal load balancers, preventing
                   unintended access to your internal load balancer through an internet gateway.

            The following attributes are supported by only Application Load Balancers:
                1. 'idle_timeout.timeout_seconds' - The idle timeout value, in seconds. The valid range is 1-4000
                   seconds. The default is 60 seconds.
                2. 'routing.http.desync_mitigation_mode' - Determines how the load balancer handles requests that might
                   pose a security risk to your application. The possible values are monitor, defensive, and strictest.
                   The default is defensive .
                3. 'routing.http.drop_invalid_header_fields.enabled' - Indicates whether HTTP headers with invalid
                   header fields are removed by load balancer(true) or routed to target(s) (false). Default is false.
                4. 'routing.http.preserve_host_header.enabled' - Indicates whether the Application Load Balancer should
                   preserve the Host header in the HTTP request and send it to the target without any change. The
                   possible values are true and false. The default is false.
                5. 'routing.http.x_amzn_tls_version_and_cipher_suite.enabled' - Indicates whether the two headers
                   (x-amzn-tls-version and x-amzn-tls-cipher-suite ), which contain information about the negotiated TLS
                   version and cipher suite, are added to the client request before sending it to the target. The
                   x-amzn-tls-version header has information about the TLS protocol version negotiated with the client,
                   and the x-amzn-tls-cipher-suite header has information about the cipher suite negotiated with the
                   client. Both headers are in OpenSSL format. The possible values for the attribute are true and false.
                   The default is false.
                6. 'routing.http.xff_client_port.enabled' - Indicates whether the X-Forwarded-For header should preserve
                   the source port that the client used to connect to the load balancer. The possible values are true
                   and false . The default is false .
                7. 'routing.http.xff_header_processing.mode' - Enables you to modify, preserve, or remove the
                   X-Forwarded-For header in the HTTP request before the Application Load Balancer sends the request to
                   the target. The possible values: append, preserve, and remove. The default is 'append'.

                   If the value is 'append', the Application Load Balancer adds the client IP address (of the last hop)
                      to the X-Forwarded-For header in the HTTP request before it sends it to targets.
                   If the value is 'preserve' the Application Load Balancer preserves the X-Forwarded-For header in the
                      HTTP request, and sends it to targets without any change.
                   If the value is 'remove', Application Load Balancer removes the X-Forwarded-For header in the HTTP
                      request before it sends it to targets.
                8. 'routing.http2.enabled' - Indicates whether HTTP/2 is enabled. The possible values are true & false.
                   The default is true . Elastic Load Balancing requires that message header names contain only
                   alphanumeric characters and hyphens.
                9. 'waf.fail_open.enabled' - Indicates whether to allow a WAF-enabled load balancer to route requests to
                   target(s) if it is unable to forward the request to Amazon Web Services WAF. The possible values are
                   true and false . The default is false .

            The following attribute is supported by Network Load Balancers and Gateway Load Balancers:
                1. 'load_balancing.cross_zone.enabled' - Indicates whether cross-zone load balancing is enabled. The
                   possible values are true and false . The default is false .
            Value (Text): The value of the attribute.
        resource_id:(Text, optional): The Amazon Resource Name (ARN) of the load balancer.

    Request Syntax:
        [load_balancer-name]:
          aws.elbv2.load_balancer.present:
          - name: 'string'
          - subnets: 'list'
          - subnet_mappings: 'list'
          - security_groups: 'list'
          - scheme: 'string'
          - lb_type: 'string',
          - ip_address_type: 'string',
          - customer_owned_ipv4_pool: 'string',
          - attributes: 'list'
          - tags:
              'string': 'string'
          - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

    . code-block:: sls

        test_load-balancer_name:
          aws.elbv2.load_balancer.present:
            - name: my-load-balancer
            - resource_id: arn:aws:elasticloadbalancing:us-west-2:123456789012:loadbalancer/app/my-load-balancer/50dc6c495c0c9188
            - tags:
                name: load-balancer-name
            - availability_zones
              - us-west-2a
            - subnets
              - subnet-15aaab61
            - security_groups
              - sg-a61988c3
            - scheme: internal

    """
    result = dict(comment=[], old_state=None, new_state=None, name=name, result=True)
    plan_state = {}
    before = None
    resource_updated = False
    if resource_id:
        before = await hub.exec.aws.elbv2.load_balancer.get(
            ctx=ctx, name=name, resource_id=resource_id
        )
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result
        result["old_state"] = copy.deepcopy(before["ret"])
        plan_state = copy.deepcopy(result["old_state"])
        update = {
            "LoadBalancerName": name,
            "Subnets": subnets,
            "SubnetMappings": subnet_mappings,
            "SecurityGroups": security_groups,
            "Scheme": scheme,
            "Tags": tags,
            "Type": lb_type,
            "IpAddressType": ip_address_type,
            "CustomerOwnedIpv4Pool": customer_owned_ipv4_pool,
            "Attributes": attributes,
        }
        update_return = await hub.tool.aws.elbv2.load_balancer.update(
            ctx=ctx,
            current_state=before["ret"],
            input_map=update,
            plan_state=plan_state,
            resource_id=resource_id,
        )
        result["comment"] = update_return["comment"]
        if not update_return["result"]:
            result["result"] = False
            return result
        resource_updated = bool(update_return["ret"])
        if resource_updated:
            if ctx.get("test", False):
                result["comment"] += list(
                    hub.tool.aws.comment_utils.would_update_comment(
                        resource_type="aws.elbv2.load_balancer", name=name
                    )
                )
            else:
                result["comment"] += list(
                    hub.tool.aws.comment_utils.update_comment(
                        resource_type="aws.elbv2.load_balancer", name=name
                    )
                )

        if tags is not None and tags != result["old_state"].get("tags"):
            # Update tags
            update_tags = await hub.tool.aws.elbv2.tag.update(
                ctx=ctx,
                resource_id=resource_id,
                old_tags=result["old_state"]["tags"],
                new_tags=tags,
            )
            result["comment"] += list(update_tags["comment"])
            result["result"] = update_tags["result"]
            resource_updated = resource_updated or bool(update_tags["ret"])
            if ctx.get("test", False) and update_tags["ret"]:
                plan_state["tags"] = update_tags["ret"]
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "subnets": subnets,
                    "subnet_mappings": subnet_mappings,
                    "security_groups": security_groups,
                    "scheme": scheme,
                    "tags": tags,
                    "lb_type": lb_type,
                    "ip_address_type": ip_address_type,
                    "customer_owned_ipv4_pool": customer_owned_ipv4_pool,
                    "attributes": attributes,
                },
            )
            result["comment"] = list(
                hub.tool.aws.comment_utils.would_create_comment(
                    resource_type="aws.elbv2.load_balancer", name=name
                )
            )
            return result
        ret = await hub.exec.boto3.client.elbv2.create_load_balancer(
            ctx,
            Name=name,
            Subnets=subnets,
            SubnetMappings=subnet_mappings,
            SecurityGroups=security_groups,
            Scheme=scheme,
            Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags),
            Type=lb_type,
            IpAddressType=ip_address_type,
            CustomerOwnedIpv4Pool=customer_owned_ipv4_pool,
        )
        if not ret["result"] and not ret["ret"]:
            result["comment"] = list(ret["comment"])
            result["result"] = False
            return result
        resource_id = (ret["ret"]["LoadBalancers"][0]).get("LoadBalancerArn")
        result["comment"] = list(
            hub.tool.aws.comment_utils.create_comment(
                resource_type="aws.elbv2.load_balancer", name=name
            )
        )
        if attributes:
            attributes_update = (
                await hub.exec.boto3.client.elbv2.modify_load_balancer_attributes(
                    ctx,
                    LoadBalancerArn=resource_id,
                    Attributes=attributes,
                )
            )
            if not attributes_update["result"]:
                result["comment"] += list(attributes_update["comment"])
                result["result"] = False
                return result
            result["comment"].append("Modified Attributes.")

    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif not (before and before["result"]) or resource_updated:
        after = await hub.exec.aws.elbv2.load_balancer.get(
            ctx=ctx, name=name, resource_id=resource_id
        )
        if not after["result"] or not after["ret"]:
            result["result"] = False
            result["comment"] += after["comment"]
            return result
        result["new_state"] = copy.deepcopy(after["ret"])
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])

    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    1. Deletes the specified Application Load Balancer, Network Load Balancer, or Gateway Load Balancer. Deleting a
       load balancer also deletes its listeners.
    2. You can't delete a load balancer if deletion protection is enabled. If the load balancer does not exist or has
       already been deleted, the call succeeds.
    3. Deleting a load balancer does not affect its registered targets. For example, your EC2 instances continue to run
       and are still registered to their target groups. If you no longer need these EC2 instances, you can stop or
       terminate them.

    Args:
        name(Text): Idem name of the load balancer.
        resource_id(Text, optional): The Amazon Resource Name (ARN) of the load balancer.

    Request Syntax:
        [load_balancer-name]:
          aws.elbv2.load_balancer.absent:
          - name: 'string'
          - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            test_load-balancer_name:
                aws.elbv2.load_balancer.absent:
                  - name: my-load-balancer
                  - resource_id: arn:aws:elasticloadbalancing:us-west-2:123456789012:loadbalancer/app/my-load-balancer/50dc6c495c0c9188

    """
    result = dict(comment=[], old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = list(
            hub.tool.aws.comment_utils.already_absent_comment(
                resource_type="aws.elbv2.load_balancer", name=name
            )
        )
        return result
    before = await hub.exec.aws.elbv2.load_balancer.get(
        ctx=ctx, name=name, resource_id=resource_id
    )
    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result

    if not before["ret"]:
        result["comment"] = list(
            hub.tool.aws.comment_utils.already_absent_comment(
                resource_type="aws.elbv2.load_balancer", name=name
            )
        )
    else:
        result["old_state"] = before["ret"]
        if ctx.get("test", False):
            result["comment"] = list(
                hub.tool.aws.comment_utils.would_delete_comment(
                    resource_type="aws.elbv2.load_balancer", name=name
                )
            )
            return result
        else:
            ret = await hub.exec.boto3.client.elbv2.delete_load_balancer(
                ctx, LoadBalancerArn=resource_id
            )
            if not ret["result"]:
                result["comment"] += list(ret["comment"])
                result["result"] = False
                return result
            result["comment"] = list(
                hub.tool.aws.comment_utils.delete_comment(
                    resource_type="aws.elbv2.load_balancer", name=name
                )
            )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function
    Lists out all load balancers.

    Returns:
        Dict[str, Any]

    Examples:

    .. code-block:: bash

        $ idem describe aws.elbv2.load_balancer
    """
    result = {}
    ret = await hub.exec.boto3.client.elbv2.describe_load_balancers(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe aws.elbv2.load_balancer: {ret['comment']}")
        return result
    if ret["ret"].get("LoadBalancers"):
        for load_balancer in ret["ret"].get("LoadBalancers"):
            tags_ret = await hub.exec.boto3.client.elbv2.describe_tags(
                ctx, ResourceArns=[load_balancer.get("LoadBalancerArn")]
            )
            if not tags_ret["result"]:
                # Error fetching tags. If fetching tags itself fails, just skip this and continue. Since tags are
                # optional, they need not be associated with all load balancers. But API needs to succeed.
                hub.log.warning(
                    f"Failed listing tags for aws.elbv2.load_balancer '{load_balancer.get('LoadBalancerArn')}'"
                    f"Describe will skip this aws.elbv2.load_balancer and continue. "
                )
                continue

            tags = []
            if tags_ret.get("ret") and tags_ret.get("ret")["TagDescriptions"]:
                tags = (tags_ret.get("ret").get("TagDescriptions")[0]).get("Tags")

            attributes_ret = (
                await hub.exec.boto3.client.elbv2.describe_load_balancer_attributes(
                    ctx, LoadBalancerArn=load_balancer.get("LoadBalancerArn")
                )
            )
            if not attributes_ret["result"]:
                # Error fetching attributes (Attributes are optional. Not every load_balancer needs to have them.)
                hub.log.warning(
                    f"Failed listing attributes for aws.elbv2.load_balancer '{load_balancer.get('LoadBalancerArn')}'"
                    f"Describe will skip this aws.elbv2.load_balancer and continue. "
                )
                continue

            attributes = []
            if attributes_ret.get("ret") and attributes_ret.get("ret").get(
                "Attributes"
            ):
                attributes = attributes_ret["ret"].get("Attributes")

            resource_converted = hub.tool.aws.elbv2.conversion_utils.convert_raw_load_balancer_to_present(
                raw_resource=load_balancer,
                idem_resource_name=load_balancer.get("LoadBalancerName"),
                tags=hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags),
                attributes=attributes,
            )
            result[resource_converted["resource_id"]] = {
                "aws.elbv2.load_balancer.present": [
                    {parameter_key: parameter_value}
                    for parameter_key, parameter_value in resource_converted.items()
                ]
            }
    return result



from typing import Any
from typing import Dict
from typing import List


async def update_elasticache_parameters(
    hub,
    ctx,
    resource_id: str,
    old_parameters: List[Dict[str, Any]],
    parameter_name_values: List[Dict[str, Any]],
):
    """

    Args:
        hub:
        ctx:
        resource_id(Text): aws resource name
        old_parameters(List): list of old parameters
        parameter_name_values(List): list of required parameters

    Returns:
        {"result": True|False, "comment": "A tuple", "is_updated": True|False}

    """
    result = dict(comment=(), result=True, ret=None, is_updated=False)
    try:
        parameters_to_update = hub.tool.aws.elasticache.elasticache_utils.get_updated_cache_parameter_group(
            old_parameters, parameter_name_values
        )

        if parameters_to_update:
            n = len(parameters_to_update)

            # Can modify up to 20 parameters in a single request
            for i in range(0, n, 20):
                final_range = n if i + 20 > n else 20
                update_ret = await hub.exec.boto3.client.elasticache.modify_cache_parameter_group(
                    ctx=ctx,
                    CacheParameterGroupName=resource_id,
                    ParameterNameValues=parameters_to_update[i : i + final_range],
                )
                if not update_ret["result"]:
                    result["comment"] = update_ret["comment"]
                    result["result"] = False
                    return result
            if n > 0:
                result["is_updated"] = True
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
        return result

    return result



from typing import Any
from typing import Dict

__func_alias__ = {"list_": "list"}


async def get(hub, ctx, name, resource_id: str):
    r"""
    Provides details about a specific hosted_zone as a data-source.

    Args:
        name(Text): The name of the Idem state.
        resource_id(Text): AWS Hosted Zone id to identify the resource.
    """
    result = dict(comment=[], ret=None, result=True)

    ret = await hub.exec.boto3.client.route53.get_hosted_zone(ctx, Id=resource_id)

    if not ret["result"]:
        if "NoSuchHostedZone" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.route53.hosted_zone", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    tags_ret = await hub.exec.boto3.client.route53.list_tags_for_resource(
        ctx, ResourceType="hostedzone", ResourceId=resource_id
    )
    tags = None
    if tags_ret["result"]:
        tags = tags_ret["ret"].get("ResourceTagSet").get("Tags", [])
    else:
        result["result"] = False
        result["comment"] = tags_ret["comment"]
        return result

    result[
        "ret"
    ] = hub.tool.aws.route53.conversion_utils.convert_raw_hosted_zone_to_present(
        raw_resource=ret,
        idem_resource_name=name,
        tags=tags,
    )

    return result


async def list_(
    hub,
    ctx,
    hosted_zone_name: str = None,
    private_zone: bool = None,
    vpc_id: str = None,
    tags: Dict[str, Any] = None,
):
    """
    Args:
        hosted_zone_name(Text, Optional): The domain name of hosted_zone
        private_zone(bool, Optional): Bool argument to specify a private hosted_zone. One of the filter option for hosted_zone
        vpc_id(Text, Optional):The vpc_id associated with the hosted_zone.One of the filter option for hosted_zone
        tags(Dict, Optional): Tags of the hosted_zone. One of the filter option for hosted_zone
    """
    result = dict(comment=[], ret=None, result=True)

    hosted_zones = []
    ret = await hub.tool.aws.route53.hosted_zone_utils.get_all_hosted_zones(ctx)
    if not ret["result"]:
        result["result"] = ret["result"]
        result["comment"] = ret["comment"]
        return result

    filtered_ret = hub.tool.aws.route53.hosted_zone_utils.get_hosted_zones_with_filters(
        raw_hosted_zones=ret["ret"],
        hosted_zone_name=hosted_zone_name,
        private_zone=private_zone,
        vpc_id=vpc_id,
        tags=tags,
    )
    result["comment"] = filtered_ret["comment"]
    if not filtered_ret["ret"]:
        result["result"] = filtered_ret["result"]
        return result

    if filtered_ret["ret"] is None:
        hub.tool.aws.comment_utils.list_empty_comment(
            resource_type="aws.route53.hosted_zone", name=hosted_zone_name
        )
        result["result"] = filtered_ret["result"]
        return result

    for raw_hosted_zone in filtered_ret["ret"]:
        name = raw_hosted_zone["ret"]["HostedZone"].get("Name")
        resource_id = raw_hosted_zone["ret"]["HostedZone"].get("Id").split("/")[-1]
        hosted_zone = await hub.exec.aws.route53.hosted_zone.get(
            ctx, name=name, resource_id=resource_id
        )
        if not hosted_zone["result"]:
            result["comment"] = hosted_zone["comment"]
            result["result"] = False
            return result

        hosted_zones.append(hosted_zone["ret"])

    result["ret"] = hosted_zones
    return result



import copy
from typing import Any
from typing import Dict
from typing import List


async def search_raw(
    hub, ctx, name, filters: List = None, resource_id: str = None, default: bool = None
) -> Dict:
    """
    Fetch one or more subnets from AWS. The return will be in the same format as what the boto3 api returns.

    Args:
        name(string): The name of the Idem state.
        resource_id(string, optional): AWS VPC id to identify the resource.
        default(bool, optional): Indicate whether the VPC is the default VPC.
        filters(list, optional): One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
         https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_vpcs

    Returns:
        {"result": True|False, "comment": A message List, "ret": Dict}

    """
    result = dict(comment=[], ret=None, result=True)
    syntax_validation = hub.tool.aws.search_utils.search_filter_syntax_validation(
        filters=filters
    )
    if not syntax_validation["result"]:
        result["result"] = False
        result["comment"] = list(syntax_validation["comment"])
        return result
    boto3_filter = hub.tool.aws.search_utils.convert_search_filter_to_boto3(
        filters=filters
    )
    if default is not None:
        boto3_filter.append({"Name": "is-default", "Values": [default]})
    ret = await hub.exec.boto3.client.ec2.describe_vpcs(
        ctx,
        Filters=boto3_filter,
        VpcIds=[resource_id] if resource_id else None,
    )
    result["result"] = ret["result"]
    result["comment"] = list(ret["comment"])
    result["ret"] = ret["ret"]
    return result


async def update_cidr_blocks(
    hub,
    ctx,
    vpc_id: str,
    old_ipv4_cidr_blocks: List[Dict[str, Any]],
    old_ipv6_cidr_blocks: List[Dict[str, Any]],
    new_ipv4_cidr_blocks: List[Dict[str, Any]],
    new_ipv6_cidr_blocks: List[Dict[str, Any]],
):
    """
    Update associated cidr blocks of a vpc. This function compares the existing(old) cidr blocks and the
    new cidr blocks. Cidr blocks that are in the new_cidr_blocks but not in the old_cidr_blocks will be associated to
    vpc. Cidr blocks that are in the old_cidr_blocks but not in the new_cidr_blocks will be disassociated from vpc.

    Args:
        vpc_id: The AWS resource id of the existing vpc
        old_ipv4_cidr_blocks: The ipv4 cidr blocks on the existing vpc
        old_ipv6_cidr_blocks: The ipv6 cidr blocks on the existing vpc
        new_ipv4_cidr_blocks: The expected ipv4 cidr blocks on the existing vpc
        new_ipv6_cidr_blocks: The expected ipv6 cidr blocks on the existing vpc

    Returns:
        {"result": True|False, "comment": A message Tuple, "ret": Dict}
        The ret contains the updated values. It contains empty dict if no parameter has been updated.

    """
    # If a block is None, we'll skip updating such cidr block.
    if new_ipv4_cidr_blocks is None:
        new_ipv4_cidr_blocks = old_ipv4_cidr_blocks
    if new_ipv6_cidr_blocks is None:
        new_ipv6_cidr_blocks = old_ipv6_cidr_blocks
    result = dict(comment=(), result=True, ret={})
    ipv4_cidr_block_map = {
        cidr_block.get("CidrBlock"): cidr_block for cidr_block in old_ipv4_cidr_blocks
    }
    ipv6_cidr_block_map = {
        cidr_block.get("Ipv6CidrBlock"): cidr_block
        for cidr_block in old_ipv6_cidr_blocks
    }
    cidr_blocks_to_add = []
    ipv4_cidr_block_results = copy.deepcopy(ipv4_cidr_block_map)
    ipv6_cidr_block_results = copy.deepcopy(ipv6_cidr_block_map)
    for cidr_block_set in new_ipv4_cidr_blocks:
        if cidr_block_set.get("CidrBlock") not in ipv4_cidr_block_map:
            cidr_blocks_to_add.append(
                hub.tool.aws.network_utils.generate_cidr_request_payload_for_vpc(
                    cidr_block_set, "ipv4"
                )
            )
            ipv4_cidr_block_results[cidr_block_set.get("CidrBlock")] = cidr_block_set
        else:
            ipv4_cidr_block_map.pop(cidr_block_set.get("CidrBlock"), None)
    for ipv6_cidr_block_set in new_ipv6_cidr_blocks:
        if ipv6_cidr_block_set.get("Ipv6CidrBlock") not in ipv6_cidr_block_map:
            cidr_blocks_to_add.append(
                hub.tool.aws.network_utils.generate_cidr_request_payload_for_vpc(
                    ipv6_cidr_block_set, "ipv6"
                )
            )
            ipv6_cidr_block_results[
                ipv6_cidr_block_set.get("Ipv6CidrBlock")
            ] = ipv6_cidr_block_set
        else:
            ipv6_cidr_block_map.pop(ipv6_cidr_block_set.get("Ipv6CidrBlock"), None)
            ipv6_cidr_block_results.pop(ipv6_cidr_block_set.get("Ipv6CidrBlock"), None)
    cidr_blocks_to_remove = list(ipv4_cidr_block_map.values()) + list(
        ipv6_cidr_block_map.values()
    )
    for cidr_to_remove in ipv4_cidr_block_map:
        ipv4_cidr_block_results.pop(cidr_to_remove)
    for cidr_to_remove in ipv6_cidr_block_map:
        ipv6_cidr_block_results.pop(cidr_to_remove)
    if (not cidr_blocks_to_remove) and (not cidr_blocks_to_add):
        return result
    if not ctx.get("test", False):
        for cidr_block in cidr_blocks_to_remove:
            ret = await hub.exec.boto3.client.ec2.disassociate_vpc_cidr_block(
                ctx, AssociationId=cidr_block.get("AssociationId")
            )
            if not ret.get("result"):
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
        for request_payload in cidr_blocks_to_add:
            ret = await hub.exec.boto3.client.ec2.associate_vpc_cidr_block(
                ctx, VpcId=vpc_id, **request_payload
            )
            if not ret.get("result"):
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
    result["comment"] = (f"Update cidr_blocks",)
    result["ret"] = {
        "cidr_block_association_set": list(ipv4_cidr_block_results.values()),
        "ipv6_cidr_block_association_set": list(ipv6_cidr_block_results.values()),
    }
    return result


async def update_vpc_attributes(
    hub,
    ctx,
    enable_dns_hostnames,
    enable_dns_support,
    old_dns_hostnames,
    old_dns_support,
    resource_id,
):
    """
    Update enable_dns_hostnames and enable_dns_support of a vpc.

    Returns:
        {"result": True|False, "comment": A message Tuple, "ret": Dict}
        The ret contains the updated values. It contains empty dict if no parameter has been updated.
    """
    result = dict(comment=(), result=True, ret={})
    if enable_dns_support is not None and enable_dns_support != old_dns_support:
        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.ec2.modify_vpc_attribute(
                ctx,
                EnableDnsSupport={"Value": enable_dns_support},
                VpcId=resource_id,
            )
            if not update_ret["result"]:
                result["result"] = False
                result["comment"] = update_ret["comment"]
        if result["result"]:
            result["ret"].update({"enable_dns_support": enable_dns_support})
    if enable_dns_hostnames is not None and enable_dns_hostnames != old_dns_hostnames:
        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.ec2.modify_vpc_attribute(
                ctx,
                EnableDnsHostnames={"Value": enable_dns_hostnames},
                VpcId=resource_id,
            )
            if not update_ret["result"]:
                result["result"] = False
                result["comment"] += update_ret["comment"]
        if result["result"]:
            result["ret"].update({"enable_dns_hostnames": enable_dns_hostnames})
    return result



"""Exec module for managing s3 bucket."""
from typing import Dict

__func_alias__ = {"list_": "list"}


async def get(hub, ctx, name: str, resource_id: str) -> Dict:
    """Returns the S3 bucket.
    Args:
        name(str): The name of the Idem state.
        resource_id(str): AWS S3 bucket name.
    """
    result = dict(comment=[], ret=None, result=True)
    resource = hub.tool.boto3.resource.create(ctx, "s3", "Bucket", resource_id)
    ret = await hub.tool.boto3.resource.describe(resource)
    if not ret:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.s3.bucket", name=name
            )
        )
        return result

    result["ret"] = await hub.tool.aws.s3.conversion_utils.convert_raw_s3_to_present(
        ctx, idem_resource_name=resource_id
    )
    return result


async def list_(
    hub,
    ctx,
    name,
) -> Dict:
    """Fetch a list of bucket from AWS.

    The function returns empty list when no resource is found.

    Args:
        name(str):
            The name of the Idem state.
    """
    result = dict(comment=[], ret=[], result=True)
    ret = await hub.exec.boto3.client.s3.list_buckets(ctx)
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["Buckets"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.s3.bucket", name=name
            )
        )
        return result
    for bucket in ret["ret"]["Buckets"]:
        result["ret"].append(
            await hub.tool.aws.s3.conversion_utils.convert_raw_s3_to_present(
                ctx=ctx, idem_resource_name=bucket["Name"]
            )
        )
    return result



import copy
from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List


async def update_tags(
    hub,
    ctx,
    resource_id,
    old_tags: List[Dict[str, Any]],
    new_tags: List[Dict[str, Any]],
):
    """
    Update tags of AWS Cost Category
    Args:
        hub:
        ctx:
        resource_id: aws resource id
        old_tags: list of old tags
        new_tags: list of new tags

    Returns:
        {"result": True|False, "comment": A message Tuple, "ret": None}

    """
    tags_to_add = []
    tags_to_remove = []
    old_tags_map = {tag.get("Key"): tag for tag in old_tags or []}
    tags_result = copy.deepcopy(old_tags_map)
    if new_tags is not None:
        for tag in new_tags:
            if tag.get("Key") in old_tags_map:
                if tag.get("Value") != old_tags_map.get(tag.get("Key")).get("Value"):
                    tags_to_add.append(tag)
                del old_tags_map[tag.get("Key")]
            else:
                tags_to_add.append(tag)
        tags_to_remove = [tag.get("Key") for tag in old_tags_map.values()]
    result = dict(comment=(), result=True, ret=None)
    if (not tags_to_remove) and (not tags_to_add):
        return result
    if tags_to_remove:
        if not ctx.get("test", False):
            delete_ret = await hub.exec.boto3.client.ce.untag_resource(
                ctx, ResourceArn=resource_id, ResourceTagKeys=tags_to_remove
            )
            if not delete_ret["result"]:
                result["comment"] = delete_ret["comment"]
                result["result"] = False
                return result
        [tags_result.pop(key, None) for key in tags_to_remove]
    if tags_to_add:
        if not ctx.get("test", False):
            add_ret = await hub.exec.boto3.client.ce.tag_resource(
                ctx, ResourceArn=resource_id, ResourceTags=tags_to_add
            )
            if not add_ret["result"]:
                result["comment"] = add_ret["comment"]
                result["result"] = False
                return result
    result["ret"] = {"tags": list(tags_result.values()) + tags_to_add}
    result["comment"] = (f"Update tags: Add [{tags_to_add}] Remove [{tags_to_remove}]",)
    return result


async def update_cost_category_definition(
    hub,
    ctx,
    before: Dict[str, Any],
    rules: List[Dict[str, Any]],
    default_value: str,
    split_charge_rules: List[Dict[str, Any]],
):
    """
    Updates the Cost Category Definition
    Args:
        before: existing resource
        rules(List[Dict[str, Any]]): The Cost Category rules used to categorize costs.
        split_charge_rules(List[Dict[str, Any]]): The split charge rules used to allocate your charges between your Cost Category values.
        default_value(Text): The default value for the cost category.

    Returns:
        {"result": True|False, "comment": A message Tuple, "ret": None}

    """
    result = dict(comment=(), result=True, ret=None)
    update_payload = {}

    if not hub.tool.aws.state_comparison_utils.are_lists_identical(
        rules,
        before.get("Rules"),
    ):
        update_payload["Rules"] = rules

    if not hub.tool.aws.state_comparison_utils.are_lists_identical(
        split_charge_rules,
        before.get("SplitChargeRules"),
    ):
        update_payload["SplitChargeRules"] = split_charge_rules

    if default_value and default_value != before.get("DefaultValue"):
        update_payload["DefaultValue"] = default_value

    if update_payload:
        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.ce.update_cost_category_definition(
                ctx,
                CostCategoryArn=before.get("CostCategoryArn"),
                RuleVersion=before.get("RuleVersion"),
                Rules=rules,
                SplitChargeRules=split_charge_rules,
                DefaultValue=default_value,
            )
            if not update_ret["result"]:
                result["comment"] = result["comment"] + update_ret["comment"]
                result["result"] = False
                return result
        result["ret"] = {}
        result = update_result(result, update_payload)

    return result


def update_result(
    result: Dict[str, Any], update_payload: Dict[str, Any]
) -> Dict[str, Any]:
    parameters = OrderedDict(
        {
            "DefaultValue": "default_value",
            "SplitChargeRules": "split_charge_rules",
            "Rules": "rules",
        }
    )

    for raw_parameter, present_parameter in parameters.items():
        if raw_parameter in update_payload:
            result["ret"][present_parameter] = update_payload[raw_parameter]
            result["comment"] = result["comment"] + (
                f"Update {present_parameter}: {update_payload[raw_parameter]}",
            )
    return result



"""
hub.exec.boto3.client.route53.list_resource_record_sets
hub.exec.boto3.client.route53.change_resource_record_sets
"""
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    hosted_zone_id: str,
    record_type: str,
    resource_records: List[str] = None,
    alias_target: Dict = None,
    resource_id: str = None,
    ttl: int = None,
) -> Dict[str, Any]:
    r"""Creates or changes a resource record set, which contains authoritative DNS information for a specified domain name
    or subdomain name.

    Args:
        name(str):
            The name of the record. A (``.``) will be appended if not already present.

        hosted_zone_id(str):
            The ID of the hosted zone that contains the resource record sets.

        record_type(str):
            The DNS record type. For information about different record types and how data is encoded for
            them, see `Supported DNS Resource Record Types` https://docs.aws.amazon.com/Route53/latest/DeveloperGuide/ResourceRecordTypes.html
            in the Amazon Route 53 Developer Guide.

        resource_records(list, Optional):
            Information about the resource records to act upon.

        alias_target(dict, Optional):
            Alias resource record sets only: Information about the Amazon Web Services resource, such as a CloudFront
            distribution or an Amazon S3 bucket, that you want to route traffic to.

        resource_id(str, Optional):
            Composite ID for a resource record in a hosted zone. String formatted as
            <hosted_zone_id>_<record_name>_<record_type>.

        ttl(int, Optional):
            The resource record cache time to live (TTL), in seconds.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            KL1PX9DBMUY9WHB_www.example.com_AAAA:
              aws.route53.resource_record.present:
                - hosted_zone_id: /hostedzone/KL1PX9DBMUY9WHB
                - name: www.example.com.
                - record_type: AAAA
                - resource_records:
                    - 2001:0db8:85a3:0:0:8a2e:0370:7335
                    - 2001:0db8:85a3:0:0:8a2e:0370:7334
                - ttl: 300

            ZY51FUS5VYB_www.example.net_A:
              aws.route53.resource_record.present:
                - alias_target:
                    dns_name: lb1.us-east-1.elb.amazonaws.com.
                    evaluate_target_health: false
                    hosted_zone_id: Z35SXDOTRQ7X7Z
                - hosted_zone_id: /hostedzone/ZY51FUS5VYB
                - name: www.example.net.
                - record_type: A
                - resource_id: ZY51FUS5VYB_www.example.net_A
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    # name is supposed to end with '.'.
    # Because name is used when we look up the resource record we should store the canonical form in the state
    formatted_name = name if name.endswith(".") else f"{name}."
    if resource_id:
        if (
            hub.tool.aws.route53.resource_record_utils.compose_resource_id(
                hosted_zone_id, name, record_type
            )
            != resource_id
        ):
            result["comment"] = (
                f"aws.route53.resource_record resource_id {resource_id} is not composed of {hosted_zone_id}, {name} and {record_type}",
            )
            result["result"] = False
            return result

    before_response = (
        await hub.tool.aws.route53.resource_record_utils.find_resource_record(
            ctx, hosted_zone_id, formatted_name, record_type
        )
    )
    if not before_response["result"]:
        result["comment"] = before_response["comment"]
        result["result"] = False
        return result
    before = before_response.get("aws_state")

    pending_state = None
    if before:
        result[
            "old_state"
        ] = hub.tool.aws.route53.conversion_utils.convert_raw_resource_record_to_present(
            hosted_zone_id=hosted_zone_id, raw_resource=before
        )
        pending_state = hub.tool.aws.route53.resource_record_utils.patch_state(
            before, hosted_zone_id, ttl, resource_records, alias_target
        )
    else:
        pending_state = {
            "name": formatted_name,
            "hosted_zone_id": hosted_zone_id,
            "resource_id": resource_id
            if resource_id
            else hub.tool.aws.route53.resource_record_utils.compose_resource_id(
                hosted_zone_id, formatted_name, record_type
            ),
            "record_type": record_type,
        }
        if ttl:
            pending_state["ttl"] = ttl
        if resource_records:
            pending_state["resource_records"] = resource_records
        if alias_target:
            pending_state["alias_target"] = alias_target
    # action can be 'CREATE'|'DELETE'|'UPSERT' - see boto documentation for change_resource_record_sets
    action = None
    if before:
        if not hub.tool.aws.route53.resource_record_utils.same_states(
            pending_state, result["old_state"]
        ):
            result["comment"] = hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.route53.resource_record", name=formatted_name
            )
            action = "UPSERT"
        else:
            result["comment"] = (
                f"No changes would be made for aws.route53.resource_record {formatted_name}",
            )
        result["result"] = True
    else:
        result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
            resource_type="aws.route53.resource_record", name=formatted_name
        )
        action = "CREATE"
        result["result"] = True

    if ctx.get("test", False):
        result["new_state"] = pending_state
        return result

    # do the actual update
    if action:
        change_batch = (
            hub.tool.aws.route53.resource_record_utils.create_change_batch_for_update(
                action, pending_state
            )
        )
        response = await hub.exec.boto3.client.route53.change_resource_record_sets(
            ctx, HostedZoneId=hosted_zone_id, ChangeBatch=change_batch
        )
        if not response["result"]:
            result["comment"] = (
                f'Error on {action} for aws.route53.resource_record {formatted_name}: {response["comment"]}',
            )
            result["result"] = False
            return result
        result["new_state"] = pending_state
        if action == "CREATE":
            result["comment"] = hub.tool.aws.comment_utils.create_comment(
                resource_type="aws.route53.resource_record", name=formatted_name
            )
        else:
            result["comment"] = hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.route53.resource_record", name=formatted_name
            )

    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""Deletes the specified resource record

    Args:
        name(str):
            Name of the resource record. Needed because of the Idem contract but not used.

        resource_id(str, Optional):
            Composite ID for a resource record in a hosted zone. String formatted as
            <hosted_zone_id>_<record_name>_<record_type>.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            MX_record_is_absent:
              aws.route53.resource_record.absent:
                - name: www.example.com.
                - resource_id: HSHMRK8IGWBU3PU_www.example.com_MX
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.route53.resource_record", name=name
        )
        return result

    try:
        decomposed_id = (
            hub.tool.aws.route53.resource_record_utils.decompose_resource_id(
                resource_id
            )
        )
    except Exception as e:
        result["comment"] = (f"{e.__class__.__name__}: {e}",)
        result["result"] = False
        return result
    hosted_zone_id = decomposed_id["hosted_zone_id"]
    formatted_name = f"{decomposed_id['name']}"
    record_type = decomposed_id["record_type"]

    before_response = (
        await hub.tool.aws.route53.resource_record_utils.find_resource_record(
            ctx,
            hosted_zone_id,
            formatted_name,
            record_type,
        )
    )
    if not before_response["result"]:
        result["comment"] = before_response["comment"]
        result["result"] = False
        return result

    before = before_response.get("aws_state")
    if not before:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.route53.resource_record", name=formatted_name
        )
    else:
        result[
            "old_state"
        ] = hub.tool.aws.route53.conversion_utils.convert_raw_resource_record_to_present(
            hosted_zone_id=hosted_zone_id, raw_resource=before
        )
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.route53.resource_record", name=formatted_name
        )
        if ctx.get("test", False):
            return result
        # delete the resource record
        change_batch = (
            hub.tool.aws.route53.resource_record_utils.create_change_batch_for_update(
                "DELETE", result["old_state"]
            )
        )
        response = await hub.exec.boto3.client.route53.change_resource_record_sets(
            ctx,
            HostedZoneId=hosted_zone_id,
            ChangeBatch=change_batch,
        )
        if not response["result"]:
            result["comment"] = (
                f"Error deleting aws.route53.resource_record '{formatted_name}': {response['comment']}",
            )
            result["result"] = False
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.route53.resource_record", name=formatted_name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    result = {}

    hosted_zones = await hub.exec.boto3.client.route53.list_hosted_zones(ctx)
    if not hosted_zones["result"]:
        hub.log.debug(
            f"Could not describe aws.route53.resource_record. {hosted_zones['comment']}"
        )
        return result

    for hosted_zone in hosted_zones["ret"]["HostedZones"]:
        hosted_zone_id = hosted_zone.get("Id")
        ret = await hub.exec.boto3.client.route53.list_resource_record_sets(
            ctx, HostedZoneId=hosted_zone_id
        )
        if not ret["result"]:
            hub.log.debug(
                f"Could not describe aws.route53.hosted_zone {ret['comment']}"
            )
            continue

        for resource_record_set in ret["ret"]["ResourceRecordSets"]:
            resource_translated = hub.tool.aws.route53.conversion_utils.convert_raw_resource_record_to_present(
                hosted_zone_id=hosted_zone_id, raw_resource=resource_record_set
            )
            result[resource_translated["resource_id"]] = {
                "aws.route53.resource_record.present": [resource_translated]
            }

    return result
